/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include "../BlackScholes.h"
#include <spu_mfcio.h>
#include <errno.h>
#include <spu_intrinsics.h>
#include <stdio.h>
#include <simdmath/logd2.h>
#include <simdmath/expd2.h>
#include <simdmath/sqrtd2.h>
#include <simdmath/recipd2.h>
#include <mass/expf4.h>
#include <mass/logf4.h>
#include <mass/sqrtf4.h>
#include <mass/recipf4.h>
#include <alf_accel.h>

void *cps;
void *S0;
void *K;
void *r;
void *sigma;
void *T;
void *answer;

void process1024() {
   int i;
   vector unsigned long long *cpsptr;
   vector double *S0ptr;
   vector double *Kptr;
   vector double *rptr;
   vector double *sigmaptr;
   vector double *Tptr;
   vector double *answerptr;
   vector double v_zero = (vector double) {0.0, 0.0};
   vector double v_one = (vector double) {1.0, 1.0};
   vector double v_half = (vector double) {0.5, 0.5};
   vector double INV_ROOT2PI = (vector double) {0.39894228, 0.39894228};
   vector double v_funnynumber = (vector double) {0.2316419, 0.2316419};
   vector double A1 = (vector double) { 0.319381530,  0.319381530};
   vector double A2 = (vector double) {-0.356563782, -0.356563782};
   vector double A3 = (vector double) { 1.781477937,  1.781477937};
   vector double A4 = (vector double) {-1.821255978, -1.821255978};
   vector double A5 = (vector double) { 1.330274429,  1.330274429};
   vector unsigned long long v_absflag = 
      (vector unsigned long long){0x7fffffffffffffffULL, 0x7fffffffffffffffULL};
   
   cpsptr = (vector unsigned long long *) cps;
   S0ptr = (vector double *) S0;
   Kptr = (vector double *) K;
   rptr = (vector double *) r;
   sigmaptr = (vector double *) sigma;
   Tptr = (vector double *) T;
   answerptr = (vector double *) answer;
   
   for (i=0; i<128; ++i) {
      vector double T_A, K_A, sigma_A, S0_A, r_A;
      vector double T_B, K_B, sigma_B, S0_B, r_B;
      vector double T_C, K_C, sigma_C, S0_C, r_C;
      vector double T_D, K_D, sigma_D, S0_D, r_D;
      vector unsigned long long cpflag_A;
      vector unsigned long long cpflag_B;
      vector unsigned long long cpflag_C;
      vector unsigned long long cpflag_D;
      vector double r5s2T_A, lnS0K_A, ssT_A, invssT_A, exp_mrT_A;
      vector double r5s2T_B, lnS0K_B, ssT_B, invssT_B, exp_mrT_B;
      vector double r5s2T_C, lnS0K_C, ssT_C, invssT_C, exp_mrT_C;
      vector double r5s2T_D, lnS0K_D, ssT_D, invssT_D, exp_mrT_D;
      vector unsigned long long flag1_A, flag2_A;
      vector unsigned long long flag1_B, flag2_B;
      vector unsigned long long flag1_C, flag2_C;
      vector unsigned long long flag1_D, flag2_D;
      vector double d1_A, k1_A, n1_A, cand1_A, accum1_A;
      vector double d1_B, k1_B, n1_B, cand1_B, accum1_B;
      vector double d1_C, k1_C, n1_C, cand1_C, accum1_C;
      vector double d1_D, k1_D, n1_D, cand1_D, accum1_D;
      vector double d2_A, k2_A, n2_A, cand2_A, accum2_A;
      vector double d2_B, k2_B, n2_B, cand2_B, accum2_B;
      vector double d2_C, k2_C, n2_C, cand2_C, accum2_C;
      vector double d2_D, k2_D, n2_D, cand2_D, accum2_D;
      vector double c_A, p_A;
      vector double c_B, p_B;
      vector double c_C, p_C;
      vector double c_D, p_D;
      vector unsigned long long f2d = (vector unsigned long long) 
                                               {0x00000000ffffffffULL, 0x00000000ffffffffULL};

      T_A = Tptr[0];
      T_B = Tptr[1];
      T_C = Tptr[2];
      T_D = Tptr[3];
      K_A = Kptr[0];
      K_B = Kptr[1];
      K_C = Kptr[2];
      K_D = Kptr[3];
      sigma_A = sigmaptr[0];
      sigma_B = sigmaptr[1];
      sigma_C = sigmaptr[2];
      sigma_D = sigmaptr[3];
      S0_A = S0ptr[0];
      S0_B = S0ptr[1];
      S0_C = S0ptr[2];
      S0_D = S0ptr[3];
      r_A = rptr[0];
      r_B = rptr[1];
      r_C = rptr[2];
      r_D = rptr[3];
      cpflag_A = cpsptr[0];
      cpflag_B = cpsptr[1];
      cpflag_C = cpsptr[2];
      cpflag_D = cpsptr[3];
      ssT_A = _sqrtd2(T_A);
      ssT_B = _sqrtd2(T_B);
      ssT_C = _sqrtd2(T_C);
      ssT_D = _sqrtd2(T_D);
      lnS0K_A = _recipd2(K_A);
      lnS0K_B = _recipd2(K_B);
      lnS0K_C = _recipd2(K_C);
      lnS0K_D = _recipd2(K_D);
      r5s2T_A = spu_mul(sigma_A, sigma_A);
      r5s2T_B = spu_mul(sigma_B, sigma_B);
      r5s2T_C = spu_mul(sigma_C, sigma_C);
      r5s2T_D = spu_mul(sigma_D, sigma_D);
      lnS0K_A = spu_mul(S0_A, lnS0K_A); 
      lnS0K_B = spu_mul(S0_B, lnS0K_B); 
      lnS0K_C = spu_mul(S0_C, lnS0K_C); 
      lnS0K_D = spu_mul(S0_D, lnS0K_D); 
      r5s2T_A = spu_madd(r5s2T_A, v_half, r_A);
      r5s2T_B = spu_madd(r5s2T_B, v_half, r_B);
      r5s2T_C = spu_madd(r5s2T_C, v_half, r_C);
      r5s2T_D = spu_madd(r5s2T_D, v_half, r_D);
      lnS0K_A = _logd2(lnS0K_A);
      lnS0K_B = _logd2(lnS0K_B);
      lnS0K_C = _logd2(lnS0K_C);
      lnS0K_D = _logd2(lnS0K_D);
      exp_mrT_A = spu_nmsub(r_A, T_A, v_zero);
      exp_mrT_B = spu_nmsub(r_B, T_B, v_zero);
      exp_mrT_C = spu_nmsub(r_C, T_C, v_zero);
      exp_mrT_D = spu_nmsub(r_D, T_D, v_zero);
      ssT_A = spu_mul(sigma_A, ssT_A);
      ssT_B = spu_mul(sigma_B, ssT_B);
      ssT_C = spu_mul(sigma_C, ssT_C);
      ssT_D = spu_mul(sigma_D, ssT_D);
      exp_mrT_A = _expd2(exp_mrT_A);
      exp_mrT_B = _expd2(exp_mrT_B);
      exp_mrT_C = _expd2(exp_mrT_C);
      exp_mrT_D = _expd2(exp_mrT_D);
      r5s2T_A = spu_mul(r5s2T_A, T_A);
      r5s2T_B = spu_mul(r5s2T_B, T_B);
      r5s2T_C = spu_mul(r5s2T_C, T_C);
      r5s2T_D = spu_mul(r5s2T_D, T_D);
      invssT_A = _recipd2(ssT_A);
      invssT_B = _recipd2(ssT_B);
      invssT_C = _recipd2(ssT_C);
      invssT_D = _recipd2(ssT_D);
      d1_A = spu_add(lnS0K_A, r5s2T_A);
      d1_B = spu_add(lnS0K_B, r5s2T_B);
      d1_C = spu_add(lnS0K_C, r5s2T_C);
      d1_D = spu_add(lnS0K_D, r5s2T_D);
      d1_A = spu_mul(d1_A, invssT_A);
      d1_B = spu_mul(d1_B, invssT_B);
      d1_C = spu_mul(d1_C, invssT_C);
      d1_D = spu_mul(d1_D, invssT_D);
      d2_A = spu_sub(d1_A, ssT_A);
      d2_B = spu_sub(d1_B, ssT_B);
      d2_C = spu_sub(d1_C, ssT_C);
      d2_D = spu_sub(d1_D, ssT_D);
      flag1_A = (vector unsigned long long) spu_cmpgt((vector float) v_zero, (vector float) d1_A);
      flag1_B = (vector unsigned long long) spu_cmpgt((vector float) v_zero, (vector float) d1_B);
      flag1_C = (vector unsigned long long) spu_cmpgt((vector float) v_zero, (vector float) d1_C);
      flag1_D = (vector unsigned long long) spu_cmpgt((vector float) v_zero, (vector float) d1_D);
      flag2_A = (vector unsigned long long) spu_cmpgt((vector float) v_zero, (vector float) d2_A);
      flag2_B = (vector unsigned long long) spu_cmpgt((vector float) v_zero, (vector float) d2_B);
      flag2_C = (vector unsigned long long) spu_cmpgt((vector float) v_zero, (vector float) d2_C);
      flag2_D = (vector unsigned long long) spu_cmpgt((vector float) v_zero, (vector float) d2_D);
      flag1_A = spu_sel(flag1_A, spu_rlmaskqwbyte(flag1_A, -4), f2d);
      flag1_B = spu_sel(flag1_B, spu_rlmaskqwbyte(flag1_B, -4), f2d);
      flag1_C = spu_sel(flag1_C, spu_rlmaskqwbyte(flag1_C, -4), f2d);
      flag1_D = spu_sel(flag1_D, spu_rlmaskqwbyte(flag1_D, -4), f2d);
      flag2_A = spu_sel(flag2_A, spu_rlmaskqwbyte(flag2_A, -4), f2d);
      flag2_B = spu_sel(flag2_B, spu_rlmaskqwbyte(flag2_B, -4), f2d);
      flag2_C = spu_sel(flag2_C, spu_rlmaskqwbyte(flag2_C, -4), f2d);
      flag2_D = spu_sel(flag2_D, spu_rlmaskqwbyte(flag2_D, -4), f2d);
      d1_A = (vector double) spu_and((vector unsigned int) d1_A, (vector unsigned int) v_absflag);
      d1_B = (vector double) spu_and((vector unsigned int) d1_B, (vector unsigned int) v_absflag);
      d1_C = (vector double) spu_and((vector unsigned int) d1_C, (vector unsigned int) v_absflag);
      d1_D = (vector double) spu_and((vector unsigned int) d1_D, (vector unsigned int) v_absflag);
      d2_A = (vector double) spu_and((vector unsigned int) d2_A, (vector unsigned int) v_absflag);
      d2_B = (vector double) spu_and((vector unsigned int) d2_B, (vector unsigned int) v_absflag);
      d2_C = (vector double) spu_and((vector unsigned int) d2_C, (vector unsigned int) v_absflag);
      d2_D = (vector double) spu_and((vector unsigned int) d2_D, (vector unsigned int) v_absflag);
      n1_A = spu_mul(d1_A, d1_A);
      n1_B = spu_mul(d1_B, d1_B);
      n1_C = spu_mul(d1_C, d1_C);
      n1_D = spu_mul(d1_D, d1_D);
      n2_A = spu_mul(d2_A, d2_A);
      n2_B = spu_mul(d2_B, d2_B);
      n2_C = spu_mul(d2_C, d2_C);
      n2_D = spu_mul(d2_D, d2_D);
      n1_A = spu_nmsub(n1_A, v_half, v_zero);
      n1_B = spu_nmsub(n1_B, v_half, v_zero);
      n1_C = spu_nmsub(n1_C, v_half, v_zero);
      n1_D = spu_nmsub(n1_D, v_half, v_zero);
      n2_A = spu_nmsub(n2_A, v_half, v_zero);
      n2_B = spu_nmsub(n2_B, v_half, v_zero);
      n2_C = spu_nmsub(n2_C, v_half, v_zero);
      n2_D = spu_nmsub(n2_D, v_half, v_zero);
      k1_A = spu_madd(d1_A, v_funnynumber, v_one);
      k1_B = spu_madd(d1_B, v_funnynumber, v_one);
      k1_C = spu_madd(d1_C, v_funnynumber, v_one);
      k1_D = spu_madd(d1_D, v_funnynumber, v_one);
      k2_A = spu_madd(d2_A, v_funnynumber, v_one);
      k2_B = spu_madd(d2_B, v_funnynumber, v_one);
      k2_C = spu_madd(d2_C, v_funnynumber, v_one);
      k2_D = spu_madd(d2_D, v_funnynumber, v_one);
      n1_A = _expd2(n1_A);
      n1_B = _expd2(n1_B);
      n1_C = _expd2(n1_C);
      n1_D = _expd2(n1_D);
      n2_A = _expd2(n2_A);
      n2_B = _expd2(n2_B);
      n2_C = _expd2(n2_C);
      n2_D = _expd2(n2_D);
      k1_A = _recipd2(k1_A);
      k1_B = _recipd2(k1_B);
      k1_C = _recipd2(k1_C);
      k1_D = _recipd2(k1_D);
      k2_A = _recipd2(k2_A);
      k2_B = _recipd2(k2_B);
      k2_C = _recipd2(k2_C);
      k2_D = _recipd2(k2_D);
      accum1_A = spu_madd(A5, k1_A, A4);
      accum1_B = spu_madd(A5, k1_B, A4);
      accum1_C = spu_madd(A5, k1_C, A4);
      accum1_D = spu_madd(A5, k1_D, A4);
      accum2_A = spu_madd(A5, k2_A, A4);
      accum2_B = spu_madd(A5, k2_B, A4);
      accum2_C = spu_madd(A5, k2_C, A4);
      accum2_D = spu_madd(A5, k2_D, A4);
      accum1_A = spu_madd(k1_A, accum1_A, A3);
      accum1_B = spu_madd(k1_B, accum1_B, A3);
      accum1_C = spu_madd(k1_C, accum1_C, A3);
      accum1_D = spu_madd(k1_D, accum1_D, A3);
      accum2_A = spu_madd(k2_A, accum2_A, A3);
      accum2_B = spu_madd(k2_B, accum2_B, A3);
      accum2_C = spu_madd(k2_C, accum2_C, A3);
      accum2_D = spu_madd(k2_D, accum2_D, A3);
      accum1_A = spu_madd(k1_A, accum1_A, A2);
      accum1_B = spu_madd(k1_B, accum1_B, A2);
      accum1_C = spu_madd(k1_C, accum1_C, A2);
      accum1_D = spu_madd(k1_D, accum1_D, A2);
      accum2_A = spu_madd(k2_A, accum2_A, A2);
      accum2_B = spu_madd(k2_B, accum2_B, A2);
      accum2_C = spu_madd(k2_C, accum2_C, A2);
      accum2_D = spu_madd(k2_D, accum2_D, A2);
      accum1_A = spu_madd(k1_A, accum1_A, A1);
      accum1_B = spu_madd(k1_B, accum1_B, A1);
      accum1_C = spu_madd(k1_C, accum1_C, A1);
      accum1_D = spu_madd(k1_D, accum1_D, A1);
      accum2_A = spu_madd(k2_A, accum2_A, A1);
      accum2_B = spu_madd(k2_B, accum2_B, A1);
      accum2_C = spu_madd(k2_C, accum2_C, A1);
      accum2_D = spu_madd(k2_D, accum2_D, A1);
      accum1_A = spu_mul(k1_A, accum1_A);
      accum1_B = spu_mul(k1_B, accum1_B);
      accum1_C = spu_mul(k1_C, accum1_C);
      accum1_D = spu_mul(k1_D, accum1_D);
      accum2_A = spu_mul(k2_A, accum2_A);
      accum2_B = spu_mul(k2_B, accum2_B);
      accum2_C = spu_mul(k2_C, accum2_C);
      accum2_D = spu_mul(k2_D, accum2_D);
      n1_A = spu_mul(n1_A, INV_ROOT2PI);
      n1_B = spu_mul(n1_B, INV_ROOT2PI);
      n1_C = spu_mul(n1_C, INV_ROOT2PI);
      n1_D = spu_mul(n1_D, INV_ROOT2PI);
      n2_A = spu_mul(n2_A, INV_ROOT2PI);
      n2_B = spu_mul(n2_B, INV_ROOT2PI);
      n2_C = spu_mul(n2_C, INV_ROOT2PI);
      n2_D = spu_mul(n2_D, INV_ROOT2PI);
      cand1_A = spu_nmsub(n1_A, accum1_A, v_one);
      cand1_B = spu_nmsub(n1_B, accum1_B, v_one);
      cand1_C = spu_nmsub(n1_C, accum1_C, v_one);
      cand1_D = spu_nmsub(n1_D, accum1_D, v_one);
      cand2_A = spu_nmsub(n2_A, accum2_A, v_one);
      cand2_B = spu_nmsub(n2_B, accum2_B, v_one);
      cand2_C = spu_nmsub(n2_C, accum2_C, v_one);
      cand2_D = spu_nmsub(n2_D, accum2_D, v_one);
      n1_A = spu_sel(cand1_A, spu_sub(v_one, cand1_A), (vector unsigned long long) flag1_A);
      n1_B = spu_sel(cand1_B, spu_sub(v_one, cand1_B), (vector unsigned long long) flag1_B);
      n1_C = spu_sel(cand1_C, spu_sub(v_one, cand1_C), (vector unsigned long long) flag1_C);
      n1_D = spu_sel(cand1_D, spu_sub(v_one, cand1_D), (vector unsigned long long) flag1_D);
      n2_A = spu_sel(cand2_A, spu_sub(v_one, cand2_A), (vector unsigned long long) flag2_A);
      n2_B = spu_sel(cand2_B, spu_sub(v_one, cand2_B), (vector unsigned long long) flag2_B);
      n2_C = spu_sel(cand2_C, spu_sub(v_one, cand2_C), (vector unsigned long long) flag2_C);
      n2_D = spu_sel(cand2_D, spu_sub(v_one, cand2_D), (vector unsigned long long) flag2_D);
      exp_mrT_A = spu_mul(exp_mrT_A, K_A);
      exp_mrT_B = spu_mul(exp_mrT_B, K_B);
      exp_mrT_C = spu_mul(exp_mrT_C, K_C);
      exp_mrT_D = spu_mul(exp_mrT_D, K_D);
      c_A = spu_mul(exp_mrT_A, n2_A);
      c_B = spu_mul(exp_mrT_B, n2_B);
      c_C = spu_mul(exp_mrT_C, n2_C);
      c_D = spu_mul(exp_mrT_D, n2_D);
      p_A = spu_sub(exp_mrT_A, S0_A);
      p_B = spu_sub(exp_mrT_B, S0_B);
      p_C = spu_sub(exp_mrT_C, S0_C);
      p_D = spu_sub(exp_mrT_D, S0_D);
      c_A = spu_msub(S0_A, n1_A, c_A);
      c_B = spu_msub(S0_B, n1_B, c_B);
      c_C = spu_msub(S0_C, n1_C, c_C);
      c_D = spu_msub(S0_D, n1_D, c_D);
      p_A = spu_add(p_A, c_A);
      p_B = spu_add(p_B, c_B);
      p_C = spu_add(p_C, c_C);
      p_D = spu_add(p_D, c_D);

      answerptr[0] = spu_sel(p_A, c_A, (vector unsigned long long) cpflag_A);
      answerptr[1] = spu_sel(p_B, c_B, (vector unsigned long long) cpflag_B);
      answerptr[2] = spu_sel(p_C, c_C, (vector unsigned long long) cpflag_C);
      answerptr[3] = spu_sel(p_D, c_D, (vector unsigned long long) cpflag_D);
      cpsptr += 4;
      S0ptr += 4;
      Kptr += 4;
      rptr += 4;
      sigmaptr += 4;
      Tptr += 4;
      answerptr += 4;
   }
   return;
}

void process2048() {
   int i;
   vector unsigned int *cpsptr;
   vector float *S0ptr;
   vector float *Kptr;
   vector float *rptr;
   vector float *sigmaptr;
   vector float *Tptr;
   vector float *answerptr;
   vector float v_zero = (vector float) {0.0F, 0.0F, 0.0F, 0.0F};
   vector float v_one = (vector float) {1.0F, 1.0F, 1.0F, 1.0F};
   vector float v_half = (vector float) {0.5F, 0.5F, 0.5F, 0.5F};
   vector float INV_ROOT2PI = (vector float) {0.39894228F, 0.39894228F, 0.39894228F, 0.39894228F};
   vector float v_funnynumber = (vector float) {0.2316419F, 0.2316419F, 0.2316419F, 0.2316419F};
   vector float A1 = (vector float) { 0.319381530F,  0.319381530F,  0.319381530F,  0.319381530F};
   vector float A2 = (vector float) {-0.356563782F, -0.356563782F, -0.356563782F, -0.356563782F};
   vector float A3 = (vector float) { 1.781477937F,  1.781477937F,  1.781477937F,  1.781477937F};
   vector float A4 = (vector float) {-1.821255978F, -1.821255978F, -1.821255978F, -1.821255978F};
   vector float A5 = (vector float) { 1.330274429F,  1.330274429F,  1.330274429F,  1.330274429F};
   vector unsigned int v_absflag = (vector unsigned int){0x7fffffff, 0x7fffffff, 0x7fffffff, 0x7fffffff};
  
   cpsptr = (vector unsigned int *) cps;
   S0ptr = (vector float *) S0;
   Kptr = (vector float *) K;
   rptr = (vector float *) r;
   sigmaptr = (vector float *) sigma;
   Tptr = (vector float *) T;
   answerptr = (vector float *) answer;

   for (i=0; i<128; ++i) {
      vector float T_A, K_A, sigma_A, S0_A, r_A;
      vector float T_B, K_B, sigma_B, S0_B, r_B;
      vector float T_C, K_C, sigma_C, S0_C, r_C;
      vector float T_D, K_D, sigma_D, S0_D, r_D;
      vector unsigned int cpflag_A;
      vector unsigned int cpflag_B;
      vector unsigned int cpflag_C;
      vector unsigned int cpflag_D;
      vector float r5s2T_A, lnS0K_A, ssT_A, invssT_A, exp_mrT_A;
      vector float r5s2T_B, lnS0K_B, ssT_B, invssT_B, exp_mrT_B;
      vector float r5s2T_C, lnS0K_C, ssT_C, invssT_C, exp_mrT_C;
      vector float r5s2T_D, lnS0K_D, ssT_D, invssT_D, exp_mrT_D;
      vector unsigned int flag1_A, flag2_A;
      vector unsigned int flag1_B, flag2_B;
      vector unsigned int flag1_C, flag2_C;
      vector unsigned int flag1_D, flag2_D;
      vector float d1_A, k1_A, n1_A, cand1_A, accum1_A;
      vector float d1_B, k1_B, n1_B, cand1_B, accum1_B;
      vector float d1_C, k1_C, n1_C, cand1_C, accum1_C;
      vector float d1_D, k1_D, n1_D, cand1_D, accum1_D;
      vector float d2_A, k2_A, n2_A, cand2_A, accum2_A;
      vector float d2_B, k2_B, n2_B, cand2_B, accum2_B;
      vector float d2_C, k2_C, n2_C, cand2_C, accum2_C;
      vector float d2_D, k2_D, n2_D, cand2_D, accum2_D;
      vector float c_A, p_A;
      vector float c_B, p_B;
      vector float c_C, p_C;
      vector float c_D, p_D;

      T_A = Tptr[0];
      T_B = Tptr[1];
      T_C = Tptr[2];
      T_D = Tptr[3];
      K_A = Kptr[0];
      K_B = Kptr[1];
      K_C = Kptr[2];
      K_D = Kptr[3];
      sigma_A = sigmaptr[0];
      sigma_B = sigmaptr[1];
      sigma_C = sigmaptr[2];
      sigma_D = sigmaptr[3];
      S0_A = S0ptr[0];
      S0_B = S0ptr[1];
      S0_C = S0ptr[2];
      S0_D = S0ptr[3];
      r_A = rptr[0];
      r_B = rptr[1];
      r_C = rptr[2];
      r_D = rptr[3];
      cpflag_A = cpsptr[0];
      cpflag_B = cpsptr[1];
      cpflag_C = cpsptr[2];
      cpflag_D = cpsptr[3];
      ssT_A = _sqrtf4(T_A);
      ssT_B = _sqrtf4(T_B);
      ssT_C = _sqrtf4(T_C);
      ssT_D = _sqrtf4(T_D);
      lnS0K_A = _recipf4(K_A);
      lnS0K_B = _recipf4(K_B);
      lnS0K_C = _recipf4(K_C);
      lnS0K_D = _recipf4(K_D);
      r5s2T_A = spu_mul(sigma_A, sigma_A);
      r5s2T_B = spu_mul(sigma_B, sigma_B);
      r5s2T_C = spu_mul(sigma_C, sigma_C);
      r5s2T_D = spu_mul(sigma_D, sigma_D);
      lnS0K_A = spu_mul(S0_A, lnS0K_A); 
      lnS0K_B = spu_mul(S0_B, lnS0K_B); 
      lnS0K_C = spu_mul(S0_C, lnS0K_C); 
      lnS0K_D = spu_mul(S0_D, lnS0K_D); 
      r5s2T_A = spu_madd(r5s2T_A, v_half, r_A);
      r5s2T_B = spu_madd(r5s2T_B, v_half, r_B);
      r5s2T_C = spu_madd(r5s2T_C, v_half, r_C);
      r5s2T_D = spu_madd(r5s2T_D, v_half, r_D);
      lnS0K_A = _logf4(lnS0K_A);
      lnS0K_B = _logf4(lnS0K_B);
      lnS0K_C = _logf4(lnS0K_C);
      lnS0K_D = _logf4(lnS0K_D);
      exp_mrT_A = spu_nmsub(r_A, T_A, v_zero);
      exp_mrT_B = spu_nmsub(r_B, T_B, v_zero);
      exp_mrT_C = spu_nmsub(r_C, T_C, v_zero);
      exp_mrT_D = spu_nmsub(r_D, T_D, v_zero);
      ssT_A = spu_mul(sigma_A, ssT_A);
      ssT_B = spu_mul(sigma_B, ssT_B);
      ssT_C = spu_mul(sigma_C, ssT_C);
      ssT_D = spu_mul(sigma_D, ssT_D);
      exp_mrT_A = _expf4(exp_mrT_A);
      exp_mrT_B = _expf4(exp_mrT_B);
      exp_mrT_C = _expf4(exp_mrT_C);
      exp_mrT_D = _expf4(exp_mrT_D);
      r5s2T_A = spu_mul(r5s2T_A, T_A);
      r5s2T_B = spu_mul(r5s2T_B, T_B);
      r5s2T_C = spu_mul(r5s2T_C, T_C);
      r5s2T_D = spu_mul(r5s2T_D, T_D);
      invssT_A = _recipf4(ssT_A);
      invssT_B = _recipf4(ssT_B);
      invssT_C = _recipf4(ssT_C);
      invssT_D = _recipf4(ssT_D);
      d1_A = spu_add(lnS0K_A, r5s2T_A);
      d1_B = spu_add(lnS0K_B, r5s2T_B);
      d1_C = spu_add(lnS0K_C, r5s2T_C);
      d1_D = spu_add(lnS0K_D, r5s2T_D);
      d1_A = spu_mul(d1_A, invssT_A);
      d1_B = spu_mul(d1_B, invssT_B);
      d1_C = spu_mul(d1_C, invssT_C);
      d1_D = spu_mul(d1_D, invssT_D);
      d2_A = spu_sub(d1_A, ssT_A);
      d2_B = spu_sub(d1_B, ssT_B);
      d2_C = spu_sub(d1_C, ssT_C);
      d2_D = spu_sub(d1_D, ssT_D);
      flag1_A = spu_cmpgt(v_zero, d1_A);
      flag1_B = spu_cmpgt(v_zero, d1_B);
      flag1_C = spu_cmpgt(v_zero, d1_C);
      flag1_D = spu_cmpgt(v_zero, d1_D);
      flag2_A = spu_cmpgt(v_zero, d2_A);
      flag2_B = spu_cmpgt(v_zero, d2_B);
      flag2_C = spu_cmpgt(v_zero, d2_C);
      flag2_D = spu_cmpgt(v_zero, d2_D);
      d1_A = (vector float) spu_and((vector unsigned int) d1_A, v_absflag);
      d1_B = (vector float) spu_and((vector unsigned int) d1_B, v_absflag);
      d1_C = (vector float) spu_and((vector unsigned int) d1_C, v_absflag);
      d1_D = (vector float) spu_and((vector unsigned int) d1_D, v_absflag);
      d2_A = (vector float) spu_and((vector unsigned int) d2_A, v_absflag);
      d2_B = (vector float) spu_and((vector unsigned int) d2_B, v_absflag);
      d2_C = (vector float) spu_and((vector unsigned int) d2_C, v_absflag);
      d2_D = (vector float) spu_and((vector unsigned int) d2_D, v_absflag);
      n1_A = spu_mul(d1_A, d1_A);
      n1_B = spu_mul(d1_B, d1_B);
      n1_C = spu_mul(d1_C, d1_C);
      n1_D = spu_mul(d1_D, d1_D);
      n2_A = spu_mul(d2_A, d2_A);
      n2_B = spu_mul(d2_B, d2_B);
      n2_C = spu_mul(d2_C, d2_C);
      n2_D = spu_mul(d2_D, d2_D);
      n1_A = spu_nmsub(n1_A, v_half, v_zero);
      n1_B = spu_nmsub(n1_B, v_half, v_zero);
      n1_C = spu_nmsub(n1_C, v_half, v_zero);
      n1_D = spu_nmsub(n1_D, v_half, v_zero);
      n2_A = spu_nmsub(n2_A, v_half, v_zero);
      n2_B = spu_nmsub(n2_B, v_half, v_zero);
      n2_C = spu_nmsub(n2_C, v_half, v_zero);
      n2_D = spu_nmsub(n2_D, v_half, v_zero);
      k1_A = spu_madd(d1_A, v_funnynumber, v_one);
      k1_B = spu_madd(d1_B, v_funnynumber, v_one);
      k1_C = spu_madd(d1_C, v_funnynumber, v_one);
      k1_D = spu_madd(d1_D, v_funnynumber, v_one);
      k2_A = spu_madd(d2_A, v_funnynumber, v_one);
      k2_B = spu_madd(d2_B, v_funnynumber, v_one);
      k2_C = spu_madd(d2_C, v_funnynumber, v_one);
      k2_D = spu_madd(d2_D, v_funnynumber, v_one);
      n1_A = _expf4(n1_A);
      n1_B = _expf4(n1_B);
      n1_C = _expf4(n1_C);
      n1_D = _expf4(n1_D);
      n2_A = _expf4(n2_A);
      n2_B = _expf4(n2_B);
      n2_C = _expf4(n2_C);
      n2_D = _expf4(n2_D);
      k1_A = _recipf4(k1_A);
      k1_B = _recipf4(k1_B);
      k1_C = _recipf4(k1_C);
      k1_D = _recipf4(k1_D);
      k2_A = _recipf4(k2_A);
      k2_B = _recipf4(k2_B);
      k2_C = _recipf4(k2_C);
      k2_D = _recipf4(k2_D);
      accum1_A = spu_madd(A5, k1_A, A4);
      accum1_B = spu_madd(A5, k1_B, A4);
      accum1_C = spu_madd(A5, k1_C, A4);
      accum1_D = spu_madd(A5, k1_D, A4);
      accum2_A = spu_madd(A5, k2_A, A4);
      accum2_B = spu_madd(A5, k2_B, A4);
      accum2_C = spu_madd(A5, k2_C, A4);
      accum2_D = spu_madd(A5, k2_D, A4);
      accum1_A = spu_madd(k1_A, accum1_A, A3);
      accum1_B = spu_madd(k1_B, accum1_B, A3);
      accum1_C = spu_madd(k1_C, accum1_C, A3);
      accum1_D = spu_madd(k1_D, accum1_D, A3);
      accum2_A = spu_madd(k2_A, accum2_A, A3);
      accum2_B = spu_madd(k2_B, accum2_B, A3);
      accum2_C = spu_madd(k2_C, accum2_C, A3);
      accum2_D = spu_madd(k2_D, accum2_D, A3);
      accum1_A = spu_madd(k1_A, accum1_A, A2);
      accum1_B = spu_madd(k1_B, accum1_B, A2);
      accum1_C = spu_madd(k1_C, accum1_C, A2);
      accum1_D = spu_madd(k1_D, accum1_D, A2);
      accum2_A = spu_madd(k2_A, accum2_A, A2);
      accum2_B = spu_madd(k2_B, accum2_B, A2);
      accum2_C = spu_madd(k2_C, accum2_C, A2);
      accum2_D = spu_madd(k2_D, accum2_D, A2);
      accum1_A = spu_madd(k1_A, accum1_A, A1);
      accum1_B = spu_madd(k1_B, accum1_B, A1);
      accum1_C = spu_madd(k1_C, accum1_C, A1);
      accum1_D = spu_madd(k1_D, accum1_D, A1);
      accum2_A = spu_madd(k2_A, accum2_A, A1);
      accum2_B = spu_madd(k2_B, accum2_B, A1);
      accum2_C = spu_madd(k2_C, accum2_C, A1);
      accum2_D = spu_madd(k2_D, accum2_D, A1);
      accum1_A = spu_mul(k1_A, accum1_A);
      accum1_B = spu_mul(k1_B, accum1_B);
      accum1_C = spu_mul(k1_C, accum1_C);
      accum1_D = spu_mul(k1_D, accum1_D);
      accum2_A = spu_mul(k2_A, accum2_A);
      accum2_B = spu_mul(k2_B, accum2_B);
      accum2_C = spu_mul(k2_C, accum2_C);
      accum2_D = spu_mul(k2_D, accum2_D);
      n1_A = spu_mul(n1_A, INV_ROOT2PI);
      n1_B = spu_mul(n1_B, INV_ROOT2PI);
      n1_C = spu_mul(n1_C, INV_ROOT2PI);
      n1_D = spu_mul(n1_D, INV_ROOT2PI);
      n2_A = spu_mul(n2_A, INV_ROOT2PI);
      n2_B = spu_mul(n2_B, INV_ROOT2PI);
      n2_C = spu_mul(n2_C, INV_ROOT2PI);
      n2_D = spu_mul(n2_D, INV_ROOT2PI);
      cand1_A = spu_nmsub(n1_A, accum1_A, v_one);
      cand1_B = spu_nmsub(n1_B, accum1_B, v_one);
      cand1_C = spu_nmsub(n1_C, accum1_C, v_one);
      cand1_D = spu_nmsub(n1_D, accum1_D, v_one);
      cand2_A = spu_nmsub(n2_A, accum2_A, v_one);
      cand2_B = spu_nmsub(n2_B, accum2_B, v_one);
      cand2_C = spu_nmsub(n2_C, accum2_C, v_one);
      cand2_D = spu_nmsub(n2_D, accum2_D, v_one);
      n1_A = spu_sel(cand1_A, spu_sub(v_one, cand1_A), flag1_A);
      n1_B = spu_sel(cand1_B, spu_sub(v_one, cand1_B), flag1_B);
      n1_C = spu_sel(cand1_C, spu_sub(v_one, cand1_C), flag1_C);
      n1_D = spu_sel(cand1_D, spu_sub(v_one, cand1_D), flag1_D);
      n2_A = spu_sel(cand2_A, spu_sub(v_one, cand2_A), flag2_A);
      n2_B = spu_sel(cand2_B, spu_sub(v_one, cand2_B), flag2_B);
      n2_C = spu_sel(cand2_C, spu_sub(v_one, cand2_C), flag2_C);
      n2_D = spu_sel(cand2_D, spu_sub(v_one, cand2_D), flag2_D);
      exp_mrT_A = spu_mul(exp_mrT_A, K_A);
      exp_mrT_B = spu_mul(exp_mrT_B, K_B);
      exp_mrT_C = spu_mul(exp_mrT_C, K_C);
      exp_mrT_D = spu_mul(exp_mrT_D, K_D);
      c_A = spu_mul(exp_mrT_A, n2_A);
      c_B = spu_mul(exp_mrT_B, n2_B);
      c_C = spu_mul(exp_mrT_C, n2_C);
      c_D = spu_mul(exp_mrT_D, n2_D);
      p_A = spu_sub(exp_mrT_A, S0_A);
      p_B = spu_sub(exp_mrT_B, S0_B);
      p_C = spu_sub(exp_mrT_C, S0_C);
      p_D = spu_sub(exp_mrT_D, S0_D);
      c_A = spu_msub(S0_A, n1_A, c_A);
      c_B = spu_msub(S0_B, n1_B, c_B);
      c_C = spu_msub(S0_C, n1_C, c_C);
      c_D = spu_msub(S0_D, n1_D, c_D);
      p_A = spu_add(p_A, c_A);
      p_B = spu_add(p_B, c_B);
      p_C = spu_add(p_C, c_C);
      p_D = spu_add(p_D, c_D);
      answerptr[0] = spu_sel(p_A, c_A, cpflag_A);
      answerptr[1] = spu_sel(p_B, c_B, cpflag_B);
      answerptr[2] = spu_sel(p_C, c_C, cpflag_C);
      answerptr[3] = spu_sel(p_D, c_D, cpflag_D);
      cpsptr += 4;
      S0ptr += 4;
      Kptr += 4;
      rptr += 4;
      sigmaptr += 4;
      Tptr += 4;
      answerptr += 4;
   }
   return;
}

int input_list_prepare(void *p_task_context __attribute__ ((unused)),
                       void *p_parm_ctx_buffer,
                       void *p_list_entries,
                       unsigned int current_count,
                       unsigned int total_count __attribute__ ((unused)))
{
   unsigned int double_flag;
   BlackScholesPara *BSPara;
   BSPara = (BlackScholesPara *)p_parm_ctx_buffer;
   alf_data_addr64_t ea;
  
   double_flag = (unsigned int)BSPara->double_flag;

   ALF_ACCEL_DTL_BEGIN(p_list_entries, ALF_BUF_IN, 0);

   if (double_flag) {
      ea = (alf_data_addr64_t)(BSPara->cps + (current_count << 13));
      ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(unsigned long long), ALF_DATA_INT64, ea);

      ea = (alf_data_addr64_t)(BSPara->S0 + (current_count << 13));
      ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(double), ALF_DATA_DOUBLE, ea);

      ea = (alf_data_addr64_t)(BSPara->K + (current_count << 13));
      ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(double), ALF_DATA_DOUBLE, ea);

      ea = (alf_data_addr64_t)(BSPara->r + (current_count << 13));
      ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(double), ALF_DATA_DOUBLE, ea);

      ea = (alf_data_addr64_t)(BSPara->sigma + (current_count << 13));
      ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(double), ALF_DATA_DOUBLE, ea);

      ea = (alf_data_addr64_t)(BSPara->T + (current_count << 13));
      ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(double), ALF_DATA_DOUBLE, ea);
   } else {
      ea = (alf_data_addr64_t)(BSPara->cps + (current_count << 13));
      ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(unsigned int), ALF_DATA_INT32, ea);

      ea = (alf_data_addr64_t)(BSPara->S0 + (current_count << 13));
      ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(float), ALF_DATA_FLOAT, ea);

      ea = (alf_data_addr64_t)(BSPara->K + (current_count << 13));
      ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(float), ALF_DATA_FLOAT, ea);

      ea = (alf_data_addr64_t)(BSPara->r + (current_count << 13));
      ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(float), ALF_DATA_FLOAT, ea);

      ea = (alf_data_addr64_t)(BSPara->sigma + (current_count << 13));
      ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(float), ALF_DATA_FLOAT, ea);

      ea = (alf_data_addr64_t)(BSPara->T + (current_count << 13));
      ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(float), ALF_DATA_FLOAT, ea);
   }

   ALF_ACCEL_DTL_END(p_list_entries);
   return 0;
}

int output_list_prepare(void *p_task_context __attribute__ ((unused)),
                        void *p_parm_ctx_buffer,
                        void *p_list_entries,
                        unsigned int current_count,
                        unsigned int total_count __attribute__ ((unused)))
{
   unsigned int double_flag;
   BlackScholesPara *BSPara;
   BSPara = (BlackScholesPara *)p_parm_ctx_buffer;
   alf_data_addr64_t ea;

   double_flag = (unsigned int)BSPara->double_flag;

   ALF_ACCEL_DTL_BEGIN(p_list_entries, ALF_BUF_OUT, 0);

   if (double_flag) {
     ea = (alf_data_addr64_t)(BSPara->answer + (current_count << 13));
     ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(double), ALF_DATA_DOUBLE, ea);
   } else {
     ea = (alf_data_addr64_t)(BSPara->answer + (current_count << 13));
     ALF_ACCEL_DTL_ENTRY_ADD(p_list_entries, CHUNK_SIZE / sizeof(float), ALF_DATA_FLOAT, ea);
   }

   ALF_ACCEL_DTL_END(p_list_entries);
   return 0;
}

int comp_kernel(void *p_task_context __attribute__ ((unused)),
                void *p_parm_ctx_buffer,
                void *p_input_buffer,
                void *p_output_buffer,
                void *p_inout_buffer __attribute__ ((unused)),
                unsigned int current_count __attribute__ ((unused)),
                unsigned int total_count __attribute__ ((unused)))
{
   unsigned int double_flag;
   BlackScholesPara *BSPara = (BlackScholesPara *)p_parm_ctx_buffer;
 
   cps = (void *)(p_input_buffer + 0 * CHUNK_SIZE);
   S0 = (void *)(p_input_buffer + 1 * CHUNK_SIZE); 
   K = (void *)(p_input_buffer + 2 * CHUNK_SIZE);
   r = (void *)(p_input_buffer + 3 * CHUNK_SIZE); 
   sigma = (void *)(p_input_buffer + 4 * CHUNK_SIZE);
   T = (void *)(p_input_buffer + 5 * CHUNK_SIZE);
   answer = (void *)p_output_buffer;

   double_flag = (unsigned int)BSPara->double_flag;
 
   if (double_flag) 
      process1024();
   else 
      process2048();
 
   return 0;
}

ALF_ACCEL_EXPORT_API_LIST_BEGIN
   ALF_ACCEL_EXPORT_API("", comp_kernel)
   ALF_ACCEL_EXPORT_API("", input_list_prepare)
   ALF_ACCEL_EXPORT_API("", output_list_prepare)
ALF_ACCEL_EXPORT_API_LIST_END
