/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <sys/mman.h>
#include <stdio.h>
#include <pthread.h>
#include "string.h"
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <fenv.h>
#include <sys/types.h>
#include <errno.h>
//#include <malloc_align.h>
#include <math.h>
#include <sys/time.h>
#include <string.h>

#include "alf.h"
#include "fft16M.h"

#define PATH_BUF_SIZE 1024  // Max length of path to PPU image
char spu_image_path[PATH_BUF_SIZE];   // Used to hold the complete path to SPU image
char library_name[PATH_BUF_SIZE];     // Used to hold the name of spu library
char spu_image_name[] = "alf_fft_spu";
char kernel_name[] = "accel_comp_kernel";
char input_dtl_name[] = "accel_input_dtl_prepare";
char output_dtl_name[] = "accel_output_dtl_prepare";

#ifdef _ALF_PLATFORM_HYBRID_
char ppu_image_path[PATH_BUF_SIZE];    // Used to hold the complete path to PPU image
#endif

int GenDMAInputLst(alf_wb_handle_t wb_handle, FFTPARAM *myFFTPara);
int GenDMAOutputLst(alf_wb_handle_t wb_handle, FFTPARAM *myFFTPara);

/* The path of input data file */
char *mem_file = "/huge/fft_mem.bin";
char *mem_addr0 = NULL;
char *mem_addr1 = NULL;

/* globle varible */
int num_spes;
ALF_WORK_BLOCK_TYPE_T work_block_type;
unsigned int repeat_count;
int is_SPU_DMA = 1;
int use_dataset = 0;
FFTPARAM myFFTPara __attribute__((aligned(128)));
void *ar_ptr, *ai_ptr, *br_ptr, *bi_ptr;
float *ar, *ai, *br, *bi;

int GenDMAInputLst(alf_wb_handle_t wb_handle, FFTPARAM *myFFTPara)
{
  int  k;
  unsigned long long wr, wi;

  wr = myFFTPara->pInputDataArrayR;
  wi = myFFTPara->pInputDataArrayI;
  wr += (((myFFTPara->nDMA << 8) + (myFFTPara->nSPU << 5)) << 2);
  wi += (((myFFTPara->nDMA << 8) + (myFFTPara->nSPU << 5)) << 2);
  

  switch (myFFTPara->nStage)
  {
    case 1:
      alf_wb_dtl_begin(wb_handle, ALF_BUF_IN, 0);
      break;
    case 2:
      alf_wb_dtl_begin(wb_handle, ALF_BUF_OVL_IN, 0);
      break;
    case 3:
      alf_wb_dtl_begin(wb_handle, ALF_BUF_OVL_IN, 0); 
      break;
  } 
  for (k=0; k<256; ++k) { 
    alf_wb_dtl_entry_add(wb_handle, ULL2ADDR(wr), 32, ALF_DATA_FLOAT);
    wr += 262144;
    
    alf_wb_dtl_entry_add(wb_handle, ULL2ADDR(wi), 32, ALF_DATA_FLOAT);
    wi += 262144;
  }
  alf_wb_dtl_end(wb_handle);

  return 0;
}

unsigned int q_mirror_array[64] =
{
  0x00, 0x20, 0x10, 0x30,
  0x08, 0x28, 0x18, 0x38,
  0x04, 0x24, 0x14, 0x34,
  0x0c, 0x2c, 0x1c, 0x3c,
  0x02, 0x22, 0x12, 0x32,
  0x0a, 0x2a, 0x1a, 0x3a,
  0x06, 0x26, 0x16, 0x36,
  0x0e, 0x2e, 0x1e, 0x3e,
  0x01, 0x21, 0x11, 0x31,
  0x09, 0x29, 0x19, 0x39,
  0x05, 0x25, 0x15, 0x35,
  0x0d, 0x2d, 0x1d, 0x3d,
  0x03, 0x23, 0x13, 0x33,
  0x0b, 0x2b, 0x1b, 0x3b,
  0x07, 0x27, 0x17, 0x37,
  0x0f, 0x2f, 0x1f, 0x3f
};

#define fill_mirror_write_list(_shift) {                                       \
  for( k = 0; k < 64; k++)                                                     \
  {                                                                            \
    m0 = q_mirror_array[k];                                                    \
    m1 = m0 | b0;                                                              \
    m2 = m0 | b1;                                                              \
    m3 = m1 | b1;                                                              \
    m0 = m0 << _shift;                                                         \
    m1 = m1 << _shift;                                                         \
    m2 = m2 << _shift;                                                         \
    m3 = m3 << _shift;                                                         \
                                                                               \
	alf_wb_dtl_entry_add(wb_handle, ULL2ADDR(m0+wr), 32, ALF_DATA_FLOAT);  \
  	alf_wb_dtl_entry_add(wb_handle, ULL2ADDR(m0+wi), 32, ALF_DATA_FLOAT);  \
  	alf_wb_dtl_entry_add(wb_handle, ULL2ADDR(m1+wr), 32, ALF_DATA_FLOAT);  \
  	alf_wb_dtl_entry_add(wb_handle, ULL2ADDR(m1+wi), 32, ALF_DATA_FLOAT);  \
  	alf_wb_dtl_entry_add(wb_handle, ULL2ADDR(m2+wr), 32, ALF_DATA_FLOAT);  \
  	alf_wb_dtl_entry_add(wb_handle, ULL2ADDR(m2+wi), 32, ALF_DATA_FLOAT);  \
  	alf_wb_dtl_entry_add(wb_handle, ULL2ADDR(m3+wr), 32, ALF_DATA_FLOAT);  \
  	alf_wb_dtl_entry_add(wb_handle, ULL2ADDR(m3+wi), 32, ALF_DATA_FLOAT);  \
  }                                                                            \
}

int GenDMAOutputLst(alf_wb_handle_t wb_handle, FFTPARAM *myFFTPara)
{
  int  k;
  unsigned long long wr, wi;
  unsigned int b0=0x80, b1=0x40;
  unsigned int m0, m1, m2, m3;

  switch (myFFTPara->nStage)
  {
    case 1:
      wr = myFFTPara->pOutputDataArrayR;
      wi = myFFTPara->pOutputDataArrayI;
      wr += (((myFFTPara->nDMA << 16) + (myFFTPara->nSPU << 13)) << 2);
      wi += (((myFFTPara->nDMA << 16) + (myFFTPara->nSPU << 13)) << 2);

      alf_wb_dtl_begin(wb_handle, ALF_BUF_OUT, 0 );  
      for( k = 0; k < 256; k++)
      {
        alf_wb_dtl_entry_add(wb_handle, ULL2ADDR(wr), 32, ALF_DATA_FLOAT);
        wr += 128;
        
        alf_wb_dtl_entry_add(wb_handle, ULL2ADDR(wi), 32, ALF_DATA_FLOAT); 
        wi += 128;
      }
      alf_wb_dtl_end(wb_handle);
      break;
    case 2:
      wr = myFFTPara->pOutputDataArrayR;
      wi = myFFTPara->pOutputDataArrayI;
      wr += (((myFFTPara->nDMA << 16) + (myFFTPara->nSPU << 5)) << 2);
      wi += (((myFFTPara->nDMA << 16) + (myFFTPara->nSPU << 5)) << 2); 
      
      alf_wb_dtl_begin(wb_handle, ALF_BUF_OVL_OUT, 0 );
      fill_mirror_write_list(10); 
      alf_wb_dtl_end(wb_handle);                               
      break;
    case 3:
      wr = myFFTPara->pOutputDataArrayR;
      wi = myFFTPara->pOutputDataArrayI;
      wr += (((myFFTPara->nDMA << 8) + (myFFTPara->nSPU << 5)) << 2);
      wi += (((myFFTPara->nDMA << 8) + (myFFTPara->nSPU << 5)) << 2);

      alf_wb_dtl_begin(wb_handle, ALF_BUF_OVL_OUT, 0 );  
      fill_mirror_write_list(18);
     	alf_wb_dtl_end(wb_handle);                                     
      break;
  }

  return 0;
}

#define MALLOC_BIG_ARRAYS {                                  \
  /* bottom 11 bits of ar and br addrs should be 0x000 */    \
  /* bottom 11 bits of ai and bi addrs shuold be 0x400 */    \
  unsigned long long ar_raw, ai_raw, br_raw, bi_raw;         \
  ar_ptr = malloc(0x4000800);                                \
  br_ptr = malloc(0x4000800);                                \
  ai_ptr = malloc(0x4000800);                                \
  bi_ptr = malloc(0x4000800);                                \
  ar_raw = ADDR2ULL(ar_ptr);                                 \
  br_raw = ADDR2ULL(br_ptr);                                 \
  ai_raw = ADDR2ULL(ai_ptr);                                 \
  bi_raw = ADDR2ULL(bi_ptr);                                 \
  if ((ar_raw * br_raw * ai_raw * bi_raw) == 0) {            \
    printf("ERROR: unable to malloc.  Exiting...\n");        \
    return(-1);                                              \
  }                                                          \
  ar_raw += 0x3ff;                                           \
  br_raw += 0x3ff;                                           \
  ai_raw += 0x3ff;                                           \
  bi_raw += 0x3ff;                                           \
  ar_raw &= ~0x3ff;                                          \
  br_raw &= ~0x3ff;                                          \
  ai_raw &= ~0x3ff;                                          \
  bi_raw &= ~0x3ff;                                          \
  ar_raw += (ar_raw & 0x400);                                \
  br_raw += (br_raw & 0x400);                                \
  ai_raw += 0x400 - (ai_raw & 0x400);                        \
  bi_raw += 0x400 - (bi_raw & 0x400);                        \
  ar = ULL2ADDR( ar_raw);                                    \
  br = ULL2ADDR( br_raw);                                    \
  ai = ULL2ADDR( ai_raw);                                    \
  bi = ULL2ADDR( bi_raw);                                    \
}

#define FREE_BIG_ARRAYS {                                    \
  free(bi_ptr);                                              \
  free(ai_ptr);                                              \
  free(br_ptr);                                              \
  free(ar_ptr);                                              \
}

float trigfunc(int in) {
   double x;
   x = ((double) in) * 6.2831853071796 / ((double) NP);
   return ((float) (7.0 + sin(x) + cos(2*x)));
}

void Usage(char *argv[])
{
  fprintf (stderr, "Usage %s [-c cycles -n num_spes -m repeat_count -P -D]\n", argv[0]);
  fprintf (stderr, "\t-n num_spes - number of SPEs (default is 8)\n");
  fprintf (stderr, "\t-m repeat_count - number of repeats (default is 256)\n");
  fprintf (stderr, "\t-P - host partitioning (default is accelerator partitioning)\n");
  fprintf (stderr, "\t-D - use dataset (default is no dataset)\n");
  exit(-1);
}

int create_task(alf_handle_t alf_handle, alf_task_handle_t *p_task_handle,  
                int in_buf_size, int out_buf_size, int inout_buf_size)
{
    int rc;
    alf_task_desc_handle_t desc_info_handle = ALF_NULL_HANDLE; 

    rc = alf_task_desc_create(alf_handle, ALF_ACCEL_TYPE_SPE, &desc_info_handle);
    if(rc < 0) {
      fprintf(stderr, "Call alf_task_desc_create error: %d!\n", rc);
      return -1;
    }

    alf_task_desc_set_int32(desc_info_handle, ALF_TASK_DESC_WB_PARM_CTX_BUF_SIZE, sizeof(myFFTPara));
    alf_task_desc_set_int32(desc_info_handle, ALF_TASK_DESC_WB_IN_BUF_SIZE, in_buf_size);
    alf_task_desc_set_int32(desc_info_handle, ALF_TASK_DESC_WB_OUT_BUF_SIZE, out_buf_size);
    alf_task_desc_set_int32(desc_info_handle, ALF_TASK_DESC_WB_INOUT_BUF_SIZE, inout_buf_size);
    alf_task_desc_set_int32(desc_info_handle, ALF_TASK_DESC_NUM_DTL_ENTRIES, 512);
    alf_task_desc_set_int32(desc_info_handle, ALF_TASK_DESC_NUM_DTL, 2);
    alf_task_desc_set_int32(desc_info_handle, ALF_TASK_DESC_TSK_CTX_SIZE, 0);
    alf_task_desc_set_int32(desc_info_handle, ALF_TASK_DESC_MAX_STACK_SIZE, 4096);
    alf_task_desc_set_int32(desc_info_handle, ALF_TASK_DESC_PARTITION_ON_ACCEL,(unsigned int)(is_SPU_DMA));

    alf_task_desc_set_int64(desc_info_handle, ALF_TASK_DESC_ACCEL_IMAGE_REF_L, (unsigned long long)spu_image_name);
    alf_task_desc_set_int64(desc_info_handle, ALF_TASK_DESC_ACCEL_LIBRARY_REF_L, (unsigned long long)library_name);
    alf_task_desc_set_int64(desc_info_handle, ALF_TASK_DESC_ACCEL_KERNEL_REF_L, (unsigned long long)kernel_name);
    alf_task_desc_set_int64(desc_info_handle, ALF_TASK_DESC_ACCEL_INPUT_DTL_REF_L, (unsigned long long)input_dtl_name);
    alf_task_desc_set_int64(desc_info_handle, ALF_TASK_DESC_ACCEL_OUTPUT_DTL_REF_L, (unsigned long long)output_dtl_name);
  
    rc = alf_task_create(desc_info_handle, NULL, (unsigned int)num_spes, 0, 1, p_task_handle);
    if(rc < 0) {
      fprintf(stderr, "Call alf_task_create error: %d!\n", rc);
      return -1;
    }
    rc = alf_task_desc_destroy(desc_info_handle);
    if(rc < 0){
      fprintf(stderr, "Call alf_task_desc_destroy error: %d!\n", rc);
      return -1;
    }
    
    return 0;
}

int create_work_block(alf_task_handle_t task_handle, alf_wb_handle_t *p_wb_handle, int Flag, int iStage)
{
   int rc;
   int iDMA, iSPU;
	
#ifdef SINGLE_USE
    for (iDMA = 0; iDMA < DMANUM_FOR_SPU; iDMA++)
    {
      for (iSPU = 0; iSPU < 8; iSPU++)
      {
        myFFTPara.nDMA = iDMA;
        myFFTPara.nFFTFlag = Flag;
        myFFTPara.nSPU = iSPU;
        myFFTPara.is_SPU_DMA = is_SPU_DMA;
        myFFTPara.nStage = iStage;

        rc = alf_wb_create(task_handle, ALF_WB_SINGLE, 0, p_wb_handle);
        if(rc < 0) {
          fprintf(stderr, "Call alf_wb_create error: %d\n", rc);
          return -1;
        }
        
        if (is_SPU_DMA == 0)
        {
          GenDMAInputLst(*p_wb_handle, &myFFTPara);
          GenDMAOutputLst(*p_wb_handle, &myFFTPara);
        }

        alf_wb_parm_add(*p_wb_handle, (void*)(&myFFTPara.pInputDataArrayR), 1, ALF_DATA_ADDR64, 3);
        alf_wb_parm_add(*p_wb_handle, (void*)(&myFFTPara.pInputDataArrayI), 1, ALF_DATA_ADDR64, 3);
        alf_wb_parm_add(*p_wb_handle, (void*)(&myFFTPara.pOutputDataArrayR), 1, ALF_DATA_ADDR64, 3);
        alf_wb_parm_add(*p_wb_handle, (void*)(&myFFTPara.pOutputDataArrayI), 1, ALF_DATA_ADDR64, 3);
        alf_wb_parm_add(*p_wb_handle, (void*)(&myFFTPara.nFFTFlag), 1, ALF_DATA_INT32, 2); 
        alf_wb_parm_add(*p_wb_handle, (void*)(&myFFTPara.nSPU), 1, ALF_DATA_INT32, 2); 
        alf_wb_parm_add(*p_wb_handle, (void*)(&myFFTPara.nDMA), 1, ALF_DATA_INT32, 2);    
        alf_wb_parm_add(*p_wb_handle, (void*)(&myFFTPara.nStage), 1, ALF_DATA_INT16, 1); 
        alf_wb_parm_add(*p_wb_handle, (void*)(&myFFTPara.is_SPU_DMA), 1, ALF_DATA_INT16, 1);
           
        alf_wb_enqueue(*p_wb_handle);
      }
    }
#endif

#ifdef MULTI_USE   
    for (iDMA = 0; iDMA < DMANUM_FOR_SPU; iDMA += repeat_count)
    {
      for (iSPU = 0; iSPU < 8; iSPU ++)
      {
        myFFTPara.nDMA = iDMA;
        myFFTPara.nFFTFlag = Flag;
        myFFTPara.nSPU = iSPU;
        myFFTPara.is_SPU_DMA = is_SPU_DMA;
        myFFTPara.nStage = iStage;
        
        rc = alf_wb_create(task_handle, work_block_type, repeat_count, p_wb_handle);
        if(rc < 0) {
          fprintf(stderr, "Call alf_wb_create error: %d\n", rc);
          return -1;
        }
        if (is_SPU_DMA == 0)
        {
          GenDMAInputLst(*p_wb_handle, &myFFTPara);
          GenDMAOutputLst(*p_wb_handle, &myFFTPara);
        }
        alf_wb_parm_add(*p_wb_handle, (void *)(&myFFTPara.pInputDataArrayR), 1, ALF_DATA_ADDR64, 3);
        alf_wb_parm_add(*p_wb_handle, (void *)(&myFFTPara.pInputDataArrayI), 1, ALF_DATA_ADDR64, 3);
        alf_wb_parm_add(*p_wb_handle, (void *)(&myFFTPara.pOutputDataArrayR), 1, ALF_DATA_ADDR64, 3);
        alf_wb_parm_add(*p_wb_handle, (void *)(&myFFTPara.pOutputDataArrayI), 1, ALF_DATA_ADDR64, 3);
        alf_wb_parm_add(*p_wb_handle, (void *)(&myFFTPara.nFFTFlag), 1, ALF_DATA_INT32, 2); 
        alf_wb_parm_add(*p_wb_handle, (void *)(&myFFTPara.nSPU), 1, ALF_DATA_INT32, 2); 
        alf_wb_parm_add(*p_wb_handle, (void *)(&myFFTPara.nDMA), 1, ALF_DATA_INT32, 2);
        alf_wb_parm_add(*p_wb_handle, (void *)(&myFFTPara.nStage), 1, ALF_DATA_INT16, 1);             
        alf_wb_parm_add(*p_wb_handle, (void *)(&myFFTPara.is_SPU_DMA), 1, ALF_DATA_INT16, 1);
        alf_wb_enqueue(*p_wb_handle);
      }
    }
#endif	

    return 0;
}

int main(int argc, char *argv[])
{
  int i;
  int fmem;
  unsigned int nodes;
  float lo_real, hi_real, lo_imag, hi_imag;
  int ch;
  int rc;
  int ret = 0;

  alf_handle_t           alf_handle       = ALF_NULL_HANDLE;
  alf_task_handle_t      task_handle[6];  
  alf_wb_handle_t        wb_handle[6];
  void                   *config_parms = NULL;
 	alf_dataset_handle_t   dataset_handle;
  
  struct timeval now, last;
  struct timezone tz;

  num_spes = 8;
  is_SPU_DMA = 1;
  work_block_type = ALF_WB_MULTI;
  repeat_count = 256;
  nodes =0;

  while ((ch = getopt (argc, argv, "c:n:m:PD")) != -1)
  {
      switch (ch)
      {
        case 'n':
          if (sscanf(optarg, "%d", &num_spes) != 1)
            Usage(argv);
          break;
       case 'm':
          if (sscanf(optarg, "%d", &repeat_count) != 1)
            Usage(argv);
          break;
        case 'P':
          is_SPU_DMA = 0;
          break;
        case 'D':
          use_dataset = 1;
          break;
        default:
          Usage(argv);
          break;
      }
  }

  if ((num_spes < 0) || (repeat_count > 256) || ((repeat_count & (repeat_count - 1)) != 0))
    Usage(argv);

  /* set work_block_type to SINGLE and repeat_count to 1 in case of PPU side DTL generation */
  if (is_SPU_DMA == 0) {
  	work_block_type = ALF_WB_SINGLE;
    repeat_count = 1;
  }

  printf("num_spes=%d, repeat_count=%d, is_SPU_DMA=%d, use_dataset=%d\n", num_spes, repeat_count, is_SPU_DMA, use_dataset);

  /* Create a large contiguous memory buffer by allocating a large
   * page (or more). Large page memory will also reduce the TLB thrashing.
   */
  if ((fmem = open(mem_file, O_CREAT | O_RDWR, 0755)) == -1) {
    printf("WARNING: unable to open file %s (errno=%d). Using malloc heap.\n", mem_file, errno);
    MALLOC_BIG_ARRAYS
  } else {
    /* Delete file so that huge pages will get freed on program termination.
     */
    remove(mem_file);
    
    mem_addr0 = (char *) mmap(0, 0x9000000, PROT_READ | PROT_WRITE, MAP_SHARED, fmem, 0);
    
    if (mem_addr0 != MAP_FAILED) {
    
      /* touch each of these nine large pages */
      mem_addr0[0x0800000] = (char) 0;
      mem_addr0[0x1800000] = (char) 0;	
      mem_addr0[0x2800000] = (char) 0;
      mem_addr0[0x3800000] = (char) 0;
      mem_addr0[0x4800000] = (char) 0;
      mem_addr0[0x5800000] = (char) 0;
      mem_addr0[0x6800000] = (char) 0;
      mem_addr0[0x7800000] = (char) 0;
      mem_addr0[0x8800000] = (char) 0;
    
    }

    mem_addr1 = (char *) mmap(0, 0x9000000, PROT_READ | PROT_WRITE, MAP_SHARED, fmem, 0x9000000);
    
    if (mem_addr1 != MAP_FAILED) {
    
      /* touch each of these nine large pages */
      mem_addr1[0x0800000] = (char) 0;
      mem_addr1[0x1800000] = (char) 0;
      mem_addr1[0x2800000] = (char) 0;
      mem_addr1[0x3800000] = (char) 0;
      mem_addr1[0x4800000] = (char) 0;
      mem_addr1[0x5800000] = (char) 0;
      mem_addr1[0x6800000] = (char) 0;
      mem_addr1[0x7800000] = (char) 0;
      mem_addr1[0x8800000] = (char) 0;
    
    }
    
    if (mem_addr0 == MAP_FAILED || mem_addr1 == MAP_FAILED) {
    	
      printf("ERROR: unable to mmap file %s (errno=%d). Using malloc heap.\n", mem_file, errno);
      close(fmem);
      MALLOC_BIG_ARRAYS
    }
    else {
      ar = (float *) (mem_addr0+0x0000000);
      br = (float *) (mem_addr0+0x4000000);
      ai = (float *) (mem_addr1+0x0000400);
      bi = (float *) (mem_addr1+0x4000400);
    }
  }

#ifndef SIM_RUN 
  printf("big array addrs: %llx %llx %llx %llx\n", ADDR2ULL(ar), ADDR2ULL(br), ADDR2ULL(ai), ADDR2ULL(bi));
  printf("loading big array A\n"); fflush(stdout);
  
  for (i=0; i<NP; ++i) {
    if ((i&0xfffff) == 0) { printf("%x of %x done\n", i, NP); fflush(stdout); }
    ai[i] = 0.0f;
    ar[i] = trigfunc(i);
  }
#endif

  /**************************/
  /* Init task server       */
  /**************************/
  gettimeofday(&last, &tz);

#ifdef _ALF_PLATFORM_HYBRID_
  sprintf(library_name, "alf_fft_hybrid_spu64.so");
#else
  if (sizeof(void *) == 4)
    sprintf(library_name, "alf_fft_cell_spu.so");
  else
    sprintf(library_name, "alf_fft_cell_spu64.so");
#endif

  rc = alf_init(config_parms, &alf_handle);
  if ( rc < 0 )
  {
    printf("Call alf_init error: rc=%d\n", rc);
    ret = rc;
    goto exit1;
  } 

  rc = alf_query_system_info(alf_handle,ALF_QUERY_NUM_ACCEL, ALF_ACCEL_TYPE_SPE, &nodes);
  if ( rc < 0 )
  {
    printf("Call alf_query_system_info error: %d\n", rc);
    ret = rc;
    goto exit2;
  } 
  else if( nodes <= 0 )
  {
    printf("Cannot allocate spe to use: number of nodes is %d\n", nodes);
    ret = -2;
    goto exit2;
  }
  
  if ( nodes <(unsigned int)( num_spes ))
  {
    printf("Warning: ALF only find %d spes available. Set num_spes to %d\n", nodes, nodes);
    num_spes = nodes;
  }
  
  alf_num_instances_set(alf_handle,(unsigned int)( num_spes));
  
  if (use_dataset) {
    rc = alf_dataset_create(alf_handle, &dataset_handle);
    if (rc < 0) {
      fprintf(stderr, "Call alf_dataset_create error: %d\n", rc);
      ret = rc;
      goto exit2;
    }
    rc = alf_dataset_buffer_add(dataset_handle, ar, NP*sizeof(float), ALF_DATASET_READ_WRITE);
    if (rc < 0) {
      fprintf(stderr, "Call alf_dataset_buffer_add error: %d\n", rc);
      ret = rc;
      goto exit3;
    }
    rc = alf_dataset_buffer_add(dataset_handle, ai, NP*sizeof(float), ALF_DATASET_READ_WRITE);
    if (rc < 0) {
      fprintf(stderr, "Call alf_dataset_buffer_add error: %d\n", rc);
      ret = rc;
      goto exit3;
    }
    rc = alf_dataset_buffer_add(dataset_handle, br, NP*sizeof(float), ALF_DATASET_READ_WRITE);
    if (rc < 0) {
      fprintf(stderr, "Call alf_dataset_buffer_add error: %d\n", rc);
      ret = rc;
      goto exit3;
    }
    rc = alf_dataset_buffer_add(dataset_handle, bi, NP*sizeof(float), ALF_DATASET_READ_WRITE);
    if (rc < 0) {
      fprintf(stderr, "Call alf_dataset_buffer_add error: %d\n", rc);
      ret = rc;
      goto exit3;
    }
  }

/************************* Task for FFT ********************************/
  /******************************/
  /* Prepare task queues for Stage 1 */
  /******************************/  
  rc = create_task( alf_handle, &task_handle[0], 65536, 65536, 0);
  if(rc <0) {
    ret = rc;
    goto exit2;
  }

  if (use_dataset) {
    rc = alf_task_dataset_associate(task_handle[0], dataset_handle);
    if (rc < 0) {
      fprintf(stderr, "Call alf_task_dataset_associate error: %d\n", rc);
      ret = rc;
      goto exit3;
    }
  }
              
  myFFTPara.pInputDataArrayR = ADDR2ULL(ar);
  myFFTPara.pInputDataArrayI = ADDR2ULL(ai);
  myFFTPara.pOutputDataArrayR = ADDR2ULL(br);
  myFFTPara.pOutputDataArrayI = ADDR2ULL(bi);
  
  rc = create_work_block(task_handle[0], &wb_handle[0], IS_FFT, 1);
  if(rc < 0){
    ret = rc;
    goto exit3;
  }
  rc = alf_task_finalize(task_handle[0]);
  if(rc < 0){
    fprintf(stderr, "Call alf_task_finalize error: %d\n", rc);
    ret = rc;
    goto exit3;
  }

  /******************************/
  /* Prepare task queues for Stage 2 */
  /******************************/ 
  rc = create_task( alf_handle, &task_handle[1], 0, 0, 65536);
  if(rc < 0) {
    ret = rc;
    goto exit3;
  }

  if (use_dataset) {
    rc = alf_task_dataset_associate(task_handle[1], dataset_handle);
    if (rc < 0) {
      fprintf(stderr, "Call alf_task_dataset_associate error: %d\n", rc);
      ret = rc;
      goto exit3;
    }
  }

  rc = alf_task_depends_on(task_handle[1], task_handle[0]);
  if(rc < 0){
    fprintf(stderr, "Call alf_task_depends_on error: %d\n", rc);
    ret = rc;
    goto exit3;
  }
  
  myFFTPara.pInputDataArrayR = ADDR2ULL(br);
  myFFTPara.pInputDataArrayI = ADDR2ULL(bi);
  myFFTPara.pOutputDataArrayR = ADDR2ULL(ar);
  myFFTPara.pOutputDataArrayI = ADDR2ULL(ai);
  
  rc = create_work_block(task_handle[1], &wb_handle[1], IS_FFT, 2);  
  if(rc < 0){
    ret = rc;
    goto exit3;
  }
 
  rc = alf_task_finalize(task_handle[1]);
  if(rc < 0){
    fprintf(stderr, "Call alf_task_finalize error: %d\n", rc);
    ret = rc;
    goto exit3;
  }
  /******************************/
  /* Prepare task queues for Stage 3 */
  /******************************/
  rc = create_task( alf_handle, &task_handle[2], 0, 0, 65536);
  if(rc < 0) {
    ret = rc;
    goto exit2;
  }

  if (use_dataset) {
    rc = alf_task_dataset_associate(task_handle[2], dataset_handle);
    if (rc < 0) {
      fprintf(stderr, "Call alf_task_dataset_associate error: %d\n", rc);
      ret = rc;
      goto exit3;
    }
  }

  rc = alf_task_depends_on(task_handle[2], task_handle[1]); 
  if(rc < 0) {
    fprintf(stderr, "Call alf_task_depends_on error: %d\n", rc);
    ret = rc;
    goto exit3;
  }
    
  myFFTPara.pInputDataArrayR = ADDR2ULL(ar);
  myFFTPara.pInputDataArrayI = ADDR2ULL(ai);
  myFFTPara.pOutputDataArrayR = ADDR2ULL(ar);
  myFFTPara.pOutputDataArrayI = ADDR2ULL(ai);
  
  rc = create_work_block(task_handle[2], &wb_handle[2], IS_FFT, 3); 
  if(rc < 0){
    ret = rc;
    goto exit3;
  }
 
  rc = alf_task_finalize(task_handle[2]);
  if(rc < 0) {
    fprintf(stderr, "Call alf_task_finalize error: %d\n", rc);
    ret = rc;
    goto exit3;
  }

  rc = alf_task_wait(task_handle[2], -1);
  if(rc < 0) {
    fprintf(stderr, "Call alf_task_wait error: %d\n", rc);
    ret = rc;
    goto exit3;
  }
  gettimeofday(&now, &tz);
  printf("Overall passed time for FFT is %ld us ===\n",
           (long)(now.tv_sec - last.tv_sec) * 1000000 +
               (long)(now.tv_usec - last.tv_usec));

#ifdef FULLRUN
  printf("Checking frequency results...\n"); fflush(stdout);

  for (i=0; i<NP; ++i)
  {
    if (ar[i] < -0.002 || ar[i] > 0.002 || ai[i] < -0.002 || ai[i] > 0.002)
    {
      printf("a[%d] = (%10.3f, %9.3f)\n", i, ar[i], ai[i]);
    }
  }
#endif

  gettimeofday(&last, &tz);

/*********************** Task for IFFT *******************************/  
  /******************************/
  /* Prepare task queues for Stage 1 */
  /******************************/  
  rc = create_task( alf_handle, &task_handle[3], 65536, 65536, 0);
  if(rc < 0) {
    ret = rc;
    goto exit2;
  }
  
  if (use_dataset) {
    rc = alf_task_dataset_associate(task_handle[3], dataset_handle);
    if (rc < 0) {
      fprintf(stderr, "Call alf_task_dataset_associate error: %d\n", rc);
      ret = rc;
      goto exit3;
    }
  }

  myFFTPara.pInputDataArrayR = ADDR2ULL(ar);
  myFFTPara.pInputDataArrayI = ADDR2ULL(ai);
  myFFTPara.pOutputDataArrayR = ADDR2ULL(br);
  myFFTPara.pOutputDataArrayI = ADDR2ULL(bi);

  rc = create_work_block(task_handle[3], &wb_handle[3], IS_IFFT, 1);
  if(rc < 0){
    ret = rc;
    goto exit3;
  }
 
  rc = alf_task_finalize(task_handle[3]);
  if(rc < 0) {
    fprintf(stderr, "Call alf_task_finalize error: %d\n", rc);
    ret = rc;
    goto exit3;
  }

  /******************************/
  /* Prepare task queues for Stage 2 */
  /******************************/ 
  rc = create_task( alf_handle, &task_handle[4], 0, 0, 65536);
  if(rc < 0){
    ret = rc;
    goto exit2;
  }

  if (use_dataset) {
    rc = alf_task_dataset_associate(task_handle[4], dataset_handle);
    if (rc < 0) {
      fprintf(stderr, "Call alf_task_dataset_associate error: %d\n", rc);
      goto exit3;
    }
  }

  rc = alf_task_depends_on(task_handle[4], task_handle[3]);
  if(rc < 0) {
    fprintf(stderr, "Call alf_task_depends_on error: %d\n", rc);
    ret = rc;
    goto exit3;
  }
  myFFTPara.pInputDataArrayR = ADDR2ULL(br); 
  myFFTPara.pInputDataArrayI = ADDR2ULL(bi); 
  myFFTPara.pOutputDataArrayR = ADDR2ULL(ar);
  myFFTPara.pOutputDataArrayI = ADDR2ULL(ai);
  
  rc = create_work_block(task_handle[4], &wb_handle[4], IS_IFFT, 2); 
  if(rc < 0){
    ret = rc;
    goto exit3;
  }
  rc = alf_task_finalize(task_handle[4]);
  if(rc < 0) {
    fprintf(stderr, "Call alf_task_finalize error: %d\n", rc);
    ret = rc;
    goto exit3;
  }

  /******************************/
  /* Prepare task queues for Stage 3 */
  /******************************/
  rc = create_task( alf_handle, &task_handle[5], 0, 0, 65536); 
  if(rc < 0){
    ret = rc;
    goto exit2;
  }

  if (use_dataset) {
    rc = alf_task_dataset_associate(task_handle[5], dataset_handle);
    if (rc < 0) {
      fprintf(stderr, "Call alf_task_dataset_associate error: %d\n", rc);
      ret = rc;
      goto exit3;
    }
  }

  rc = alf_task_depends_on(task_handle[5], task_handle[4]); 
   if(rc < 0) {
    fprintf(stderr, "Call alf_task_depends_on error: %d\n", rc);
    ret = rc;
    goto exit3;
  }
    
  myFFTPara.pInputDataArrayR = ADDR2ULL(ar); 
  myFFTPara.pInputDataArrayI = ADDR2ULL(ai); 
  myFFTPara.pOutputDataArrayR = ADDR2ULL(ar);
  myFFTPara.pOutputDataArrayI = ADDR2ULL(ai);
  
  rc = create_work_block(task_handle[5], &wb_handle[4], IS_IFFT, 3); 
  if(rc < 0){
    ret = rc;
    goto exit3;
  }
 
  rc = alf_task_finalize(task_handle[5]);
  if(rc < 0) {
    fprintf(stderr, "Call alf_task_finalize error: %d\n", rc);
    ret = rc;
    goto exit3;
  }


  rc = alf_task_wait(task_handle[5], -1);  
  if(rc < 0) {
    fprintf(stderr, "Call alf_task_wait error: %d\n", rc);
    ret = rc;
    goto exit3;
  }

  gettimeofday(&now, &tz);

  printf("Overall passed time for IFFT is %ld us ===\n",
           (long)(now.tv_sec - last.tv_sec) * 1000000 +
               (long)(now.tv_usec - last.tv_usec));

#ifdef FULLRUN
  printf("Now checking results...\n"); fflush(stdout);

  hi_real = -100.0;
  lo_real =  100.0;
  hi_imag = -100.0;
  lo_imag =  100.0;
  for (i=0; i<NP; ++i) {
    if ((i&0xfffff) == 0) { printf("%x of %x done\n", i, NP); fflush(stdout); }
    float x;
    x = trigfunc(i);
    if (ar[i] - x > hi_real) hi_real = ar[i] - x ;
    if (ar[i] - x < lo_real) lo_real = ar[i] - x ;
    if (ai[i]     > hi_imag) hi_imag = ai[i];
    if (ai[i]     < lo_imag) lo_imag = ai[i];
  }
  fprintf(stdout, "real err*r range = %f %f\n", lo_real, hi_real);
  fprintf(stdout, "imag err*r range = %f %f\n", lo_imag, hi_imag);
  if (lo_real < -0.00003 || hi_real > 0.00003 || lo_imag < -0.00003 || hi_imag > 0.00003)
  {
    fprintf(stderr, "ERROR: range of error values too large...\n");
    fflush(stderr);
    ret = -1;
  }
#endif
  
exit3:
  if (use_dataset) {
    rc = alf_dataset_destroy(dataset_handle);
    if (rc < 0) {
      fprintf(stderr, "Call alf_dataset_destroy error: %d\n", rc);
      ret = rc;
    }
  }
  
exit2:
  alf_exit(alf_handle, ALF_EXIT_POLICY_WAIT,-1);

exit1:  
  if (fmem == -1) {
    FREE_BIG_ARRAYS
  } else {
    if (mem_addr0 == MAP_FAILED || mem_addr1 == MAP_FAILED) {
      FREE_BIG_ARRAYS
    }
  }
  
  return ret;
}
