/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <spu_mfcio.h>
#include <stdlib.h>
#include <math.h>

#include "alf_accel.h"
#include "../pi.h"

int comp_kernel(void *p_task_ctx,
                void *p_parm_ctx_buffer,
                void *p_input_buffer __attribute__ ((unused)),
                void *p_output_buffer __attribute__ ((unused)),
                void *p_inout_buffer __attribute__ ((unused)),
                unsigned int current_count __attribute__ ((unused)),
                unsigned int total_count __attribute__ ((unused)))
{
  udata_t *my_udata_ptr = (udata_t *)p_parm_ctx_buffer;
  result_t *tsk_ctx_ptr = (result_t *)p_task_ctx;

  unsigned int i;
  double x, y;
  unsigned result = 0;

  srand(my_udata_ptr->seed);
  for (i = 0; i < my_udata_ptr->iter; i++)
  {
    x = rand() * 1.0 / RAND_MAX;
    y = rand() * 1.0 / RAND_MAX;
    if (x*x + y*y <= 1.0)
      result += 1;
  }
  tsk_ctx_ptr->value += (result * 4.0 / ITERATIONS);
  tsk_ctx_ptr->wbs ++ ;

  return 0;
}

int context_setup(void *p_task_ctx)
{
  result_t *tsk_ctx_ptr = (result_t *)p_task_ctx;

  tsk_ctx_ptr->value = 0.0;
  tsk_ctx_ptr->wbs = 0;
  return 0;
}


int context_merge(void* p_dst_task_ctx, void* p_task_ctx)
{
  result_t *tsk_ctx_ptr = (result_t *)p_task_ctx;
  result_t *dst_tsk_ctx_ptr = (result_t *)p_dst_task_ctx;

  tsk_ctx_ptr->value += dst_tsk_ctx_ptr->value;
  tsk_ctx_ptr->wbs += dst_tsk_ctx_ptr->wbs;

  return 0;
}

ALF_ACCEL_EXPORT_API_LIST_BEGIN
  ALF_ACCEL_EXPORT_API("", comp_kernel);
  ALF_ACCEL_EXPORT_API("", context_setup);
  ALF_ACCEL_EXPORT_API("", context_merge);
ALF_ACCEL_EXPORT_API_LIST_END
