# --------------------------------------------------------------  
# (C)Copyright 2007,2008,                                         
# International Business Machines Corporation                     
# All Rights Reserved.                                            
#                                                                 
# Redistribution and use in source and binary forms, with or      
# without modification, are permitted provided that the           
# following conditions are met:                                   
#                                                                 
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer. 
#                                                                 
# - Redistributions in binary form must reproduce the above       
#   copyright notice, this list of conditions and the following   
#   disclaimer in the documentation and/or other materials        
#   provided with the distribution.                               
#                                                                 
# - Neither the name of IBM Corporation nor the names of its      
#   contributors may be used to endorse or promote products       
#   derived from this software without specific prior written     
#   permission.                                                   
#                                                                 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              
# --------------------------------------------------------------  
# PROLOG END TAG zYx                                              

#  Determine the architecture of the host machine
HOST_ARCH	:= $(shell uname -m)

#
# If the TARGET_CPU is defined
# then set that target and ignore all settings in alf_examples_config.mk
#
ifdef TARGET_CPU
  TARGET_$(TARGET_CPU):=1
endif

#  Determine the location of this file, and use it to
#  locate and import the configuration file
topdir := $(dir $(realpath $(lastword $(MAKEFILE_LIST))))
include $(topdir)alf_examples_config.mk

#  This define is used as a pseudo-function to ensure the
#  values for the various LINKAGE_* variables are valid.
define check_linkage
  $(if $(strip $($(1))),,
	$(error Error in $(topdir)alf_examples_config.mk:
	No value set for $(1).
	Legal values are: 'shared' and 'static'))
  $(if $(filter-out shared,$(filter-out static,$($(1)))),
	$(error Error in $(topdir)alf_examples_config.mk:
	Illegal setting for $(1): '$($(1))'.  
	Legal values are: 'shared' and 'static'))
endef

#  This define is used to add the entries in SRC_VPATH to
#  the vpath used for .c files and to populate the VPATH_SRC
#  variable with a list of all source files found
define add_src_vpath
  vpath %.c $(abspath $(1))
  VPATH_SRCS	+=	$(notdir $(wildcard $(1)/*.c))
endef

#  This define is used to add the entried in LIB_VPATH to
#  the vpaths used for .a and .so files.  Note that this path is
#  not searched during linking, but is used to determine if the
#  objects in LIB_DEPS should cause a rebuild.
define add_lib_vpath
  vpath %.so $(1)
  vpath %.a $(1)
endef


#  Call the "check_linkage" pseudo-function to make sure
#  the linkage settings are valid
$(eval $(call check_linkage,LINKAGE_ALF))
$(eval $(call check_linkage,LINKAGE_ALF_HYBRID))
$(eval $(call check_linkage,LINKAGE_DACS_HYBRID))

#ifdef PLATFORM
_PLATFORM	=	$(strip $(PLATFORM))
#endif

#  If make is doing more than processing directories, verify
#  a legal value for PLATFORM has been set.
ifneq "$(LIBRARIES_32)$(LIBRARIES_64)$(PROGRAMS)$(PROGRAM)" ""
  ifndef _PLATFORM
    $(error Error: No value supplied for PLATFORM. Legal values are: 'ppu64', 'ppu32', 'spu', and 'x86_64')
  else ifeq "$(_PLATFORM)" "ppu64"
  else ifeq "$(_PLATFORM)" "ppu32"
  else ifeq "$(_PLATFORM)" "spu"
  else ifeq "$(_PLATFORM)" "x86_64"
  else
    $(error Error: Illegal value supplied for PLATFORM ('$(PLATFORM)').  Legal values are: 'ppu64', 'ppu32', 'spu', and 'x86_64')
  endif
endif

#  Detect ALF and ALF Hybrid components
HAVE_ALF_cbea	=	$(if $(realpath $(LIBDIR_ALF_spu)/libalf.a),1,0)
HAVE_ALF		=	$(if $(findstring 1,$(HAVE_ALF_cbea)),1,0)
HAVE_HYBRID_DACS	= 	$(if $(realpath $(LIBDIR_DACS_HYBRID_x86_64)/libdacs_hybrid.so),1,0)
HAVE_HYBRID_ALF_x86_64	=	$(if $(realpath $(LIBDIR_ALF_HYBRID_x86_64)/libalf_hybrid.so),1,0)
HAVE_HYBRID_ALF_cbea	=	$(if $(realpath $(LIBDIR_ALF_HYBRID_spu)/libalf_hybrid.a),1,0)
HAVE_HYBRID_ALF		=	$(if $(findstring 1,$(HAVE_HYBRID_ALF_cbea) $(HAVE_HYBRID_ALF_x86_64)),1,0)

#  Detect toolchains
HAVE_TOOLS_x86_64	=	$(if $(filter x86_64, $(shell if [ -n "$(CC_x86_64_gnu)" ] && [ -x $(CC_x86_64_gnu) ] ; then echo x86_64 ; fi)),1,0)
HAVE_TOOLS_cbea		=	$(if $(filter spu,$(shell if [ -n "$(CC_spu_$(COMPILER))" ] && [ -x $(CC_spu_$(COMPILER)) ] ; then echo spu ; fi)),1,0)

#  If they are not explicitly set, determine sane 
#  values for BUILD_HYBRID and BUILD_CELL
BUILD_HYBRID		?=	$(or $(and $(findstring 1,$(HAVE_HYBRID_DACS)),$(findstring 1,$(HAVE_HYBRID_ALF)),1),0)
BUILD_CELL		?=	$(HAVE_ALF)

# Only build the x86_64 or the cbea directories if both the tools chain and the ALF & ALF Hybrid components are present
BUILD_x86_64		=	$(or $(and $(findstring 1,$(HAVE_TOOLS_x86_64)),$(findstring 1,$(HAVE_HYBRID_ALF_x86_64)),1),0)
BUILD_cbea			=	$(or $(and $(findstring 1,$(HAVE_TOOLS_cbea)),$(or $(findstring 1,$(HAVE_ALF_cbea)), $(findstring 1,$(HAVE_HYBRID_ALF_cbea))),1),0)

#  Use the PLATFORM variable to set correct generic paths to
#  components
INCDIR_SDK 		=	$(INCDIR_SDK_$(_PLATFORM))

INCDIR_ALF 		= 	$(INCDIR_ALF_$(_PLATFORM))
INCDIR_ALF_HYBRID 	= 	$(INCDIR_ALF_HYBRID_$(_PLATFORM)) $(INCDIR_ALF_$(PLATFORM))

LIBDIR_ALF		=	$(LIBDIR_ALF_$(_PLATFORM))
LIBDIR_ALF_HYBRID 	=	$(LIBDIR_ALF_HYBRID_$(_PLATFORM))
LIBDIR_ALF_RUNTIME	=	$(LIBDIR_ALF_RUNTIME_$(_PLATFORM))
LIBDIR_ALF_HYBRID_RUNTIME=	$(LIBDIR_ALF_HYBRID_RUNTIME_$(_PLATFORM))

LIBDIR_ALF_DEBUG		=	$(LIBDIR_ALF_DEBUG_$(_PLATFORM))
LIBDIR_ALF_HYBRID_DEBUG 	=	$(LIBDIR_ALF_HYBRID_DEBUG_$(_PLATFORM))
LIBDIR_ALF_RUNTIME_DEBUG	=	$(LIBDIR_ALF_RUNTIME_DEBUG_$(_PLATFORM))
LIBDIR_ALF_HYBRID_RUNTIME_DEBUG=	$(LIBDIR_ALF_HYBRID_RUNTIME_DEBUG_$(_PLATFORM))

LIBDIR_DACS_HYBRID	=	$(LIBDIR_DACS_HYBRID_$(_PLATFORM))
LIBDIR_DACS_HYBRID_RUNTIME=	$(LIBDIR_DACS_HYBRID_RUNTIME_$(_PLATFORM))

#  Set the link flags for ALF.  First set the paths, then set the generic 
#  flags based on the requested linkage
LDFLAGS_ALF_static  =   -L$(LIBDIR_ALF) -Wl,-Bstatic,-lalf,-Bdynamic,-lspe2,-ldl,-lpthread
LDFLAGS_ALF_shared	=	-L$(LIBDIR_ALF) -lalf
LDFLAGS_ALF_common	=
LDFLAGS_ALF		=	$(LDFLAGS_ALF_$(LINKAGE_ALF)) $(LDFLAGS_ALF_common)
LDFLAGS_ALF_spu		=	-L$(LIBDIR_ALF_spu) -lalf
LDFLAGS_ALF_LTS_spu		=	-L$(LIBDIR_ALF_spu) -lalflts

# Debug version of LDFLAGS for ALF
LDFLAGS_ALF_DEBUG_static  =   -L$(LIBDIR_ALF_DEBUG) -Wl,-Bstatic,-lalf,-Bdynamic,-lspe2,-ldl,-lpthread
LDFLAGS_ALF_DEBUG_shared	=	-L$(LIBDIR_ALF_DEBUG) -lalf
LDFLAGS_ALF_DEBUG_common	=
LDFLAGS_ALF_DEBUG		=	$(LDFLAGS_ALF_DEBUG_$(LINKAGE_ALF)) $(LDFLAGS_ALF_DEBUG_common)
LDFLAGS_ALF_DEBUG_spu		=	-L$(LIBDIR_ALF_DEBUG_spu) -lalf
LDFLAGS_ALF_LTS_DEBUG_spu		=	-L$(LIBDIR_ALF_DEBUG_spu) -lalflts

#  Set the link flags for Hybrid ALF.  First set the paths, then set the generic 
#  flags based on the requested linkage
LDFLAGS_ALF_HYBRID_static=	-L$(LIBDIR_ALF_HYBRID) -Wl,-Bstatic,-lalf_hybrid,-Bdynamic,-ldl,-lpthread
LDFLAGS_ALF_HYBRID_shared=	-L$(LIBDIR_ALF_HYBRID) -lalf_hybrid -Wl,-rpath=$(LIBDIR_ALF_HYBRID_RUNTIME) 
LDFLAGS_ALF_HYBRID_common=
LDFLAGS_ALF_HYBRID	=	$(LDFLAGS_ALF_HYBRID_$(LINKAGE_ALF_HYBRID)) $(LDFLAGS_ALF_HYBRID_common) 
LDFLAGS_ALF_HYBRID_spu	=	-L$(LIBDIR_ALF_HYBRID_spu) -lalf_hybrid
LDFLAGS_ALF_HYBRID_LTS_spu	=	-L$(LIBDIR_ALF_HYBRID_spu) -lalflts_hybrid

# Debug version of LDFLAGS for ALF Hybrid
LDFLAGS_ALF_HYBRID_DEBUG_static=	-L$(LIBDIR_ALF_HYBRID_DEBUG) -Wl,-Bstatic,-lalf_hybrid,-Bdynamic,-ldl,-lpthread
LDFLAGS_ALF_HYBRID_DEBUG_shared=	-L$(LIBDIR_ALF_HYBRID_DEBUG) -lalf_hybrid -Wl,-rpath=$(LIBDIR_ALF_HYBRID_RUNTIME_DEBUG) 
LDFLAGS_ALF_HYBRID_DEBUG_common=
LDFLAGS_ALF_HYBRID_DEBUG	=	$(LDFLAGS_ALF_HYBRID_DEBUG_$(LINKAGE_ALF_HYBRID)) $(LDFLAGS_ALF_HYBRID_DEBUG_common) 
LDFLAGS_ALF_HYBRID_DEBUG_spu	=	-L$(LIBDIR_ALF_HYBRID_DEBUG_spu) -lalf_hybrid
LDFLAGS_ALF_HYBRID_LTS_DEBUG_spu	=	-L$(LIBDIR_ALF_HYBRID_DEBUG_spu) -lalflts_hybrid

#  Set the link flags for Hybrid DaCS.  First set the paths, then set the generic 
#  flags based on the requested linkage
LDFLAGS_DACS_HYBRID_static=	-L$(LIBDIR_DACS_HYBRID) -Wl,-Bstatic,-ldacs_hybrid,-Bdynamic
LDFLAGS_DACS_HYBRID_shared=	-L$(LIBDIR_DACS_HYBRID) -ldacs_hybrid
LDFLAGS_DACS_HYBRID_common=
LDFLAGS_DACS_HYBRID	= 	$(LDFLAGS_DACS_HYBRID_$(LINKAGE_DACS_HYBRID)) $(LDFLAGS_DACS_HYBRID_common)

#  Set the compiler based on "PLATFORM" and "COMPILER" variables
#  Note that for x86_64, gnu is the only compiler
ifneq "$(_PLATFORM)" "x86_64"
  CC			=	$(CC_$(_PLATFORM)_$(COMPILER))
else
  CC			=	$(CC_x86_64_gnu)
endif

#  Add any directories which should be processed based on toolkit availibility
DIRS			+=	$(if $(filter 1,$(BUILD_cbea)),$(DIRS_cbea))
DIRS			+=	$(if $(filter 1,$(BUILD_x86_64)),$(DIRS_x86_64))
DIRS			+=	$(if $(filter 1,$(TARGET_ppc)),$(DIRS_ppc))
DIRS			+=	$(if $(filter 1,$(TARGET_ppc64)),$(DIRS_ppc64))

#  Add any directories which should be processed based on build type
DIRS			+=	$(if $(filter 1,$(BUILD_CELL)),$(DIRS_CELL))
DIRS			+=	$(if $(filter 1,$(BUILD_HYBRID)),$(DIRS_HYBRID))

#  Add the current path plus any entries in SRC_VPATH to the path
#  searched for .c files.
$(foreach path,$(SRC_VPATH) .,$(eval $(call add_src_vpath,$(path))))

#  Add the entries of LIB_VPATH to the library search path for 
#  dependancy generation
$(foreach path,$(LIB_VPATH),$(eval $(call add_lib_vpath,$(path))))

#  Combine any library targets into a single "LIBRARIES" variable
LIBRARIES		=	$(strip $(LIBRARIES_32) $(LIBRARIES_64))

#  Add PROGRAM to PROGRAMS
PROGRAMS		+=	$(PROGRAM)

#  This define is used to generate a rule for each entry in the PROGRAMS variable
define program_template

  #  If no sources are defined for a program, default it to all the implicit sources
  $(if $(SRCS_$(1)),,$(eval SRCS_$(1)=$(IMPLICIT_SRCS)))

  #  Convert the sources to objects
  OBJS_$(1) = $(SRCS_$(1):.c=.o)

  #  This is the rule for the program
  $(1): $$(OBJS_$(1)) $(LIB_DEPS) $(LIB_DEPS_$(1))
	$(CC) -o $(1) $$(OBJS_$(1)) $(LIB_PATHS) $(LDFLAGS)  $(LDFLAGS_$(1))

endef

#  Iterate through all the entries in PROGRAMS and gather all explicit sources 
#  into a single variable, EXPLICIT_SRCS
EXPLICIT_SRCS	:=	$(foreach i,$(PROGRAMS),$(SRCS_$(i)))

#  Now, strip all the explicit sources for the list of ALL sources (leaving just 
#  the implicit sources, and store in IMPLICIT_SRCS
IMPLICIT_SRCS	= 	$(or $(SRCS),$(filter-out $(EXPLICIT_SRCS),$(VPATH_SRCS)))

#  Iterate through all the entries in PROGRAMS and create the embedded obj list
EMBED32_OBJS	:=	$(foreach i,$(PROGRAMS),$(i)-embed.o)
EMBED64_OBJS	:=	$(foreach i,$(PROGRAMS),$(i)-embed64.o)

#  Form the lib paths used in compile flags
LIB_PATHS	:=	$(foreach i,$(LIB_VPATH),-L$(i))

#  Transform the list of sources (generated above) into a list of objects
OBJS            	=       $(IMPLICIT_SRCS:.c=.o) $(EXPLICIT_SRCS:.c=.o)

#  Generate a list of objects which should be removed when the "clean"
#  target is invoked.
CLEAN_TARGETS		=	$(strip $(foreach item,$(PROGRAMS) $(LIBRARIES) $(OBJS) $(PROGRAM)-embed*.o $(EMBED32_OBJS) $(EMBED64_OBJS) $(OBJS:.o=.d),$(wildcard $(item))))

#  XLC requires special handling
ifeq "$(COMPILER)" "xlc"

  #  Exclude the x86_64 platform since is is handled by gnu
  ifneq "$(_PLATFORM)" "x86_64"

    #  Add support for C++ comments ( -qcpluscmt)
    #  Add support for generation of .d files as part of compile step (-M)
    #  Add support for generation of inline "alloca" calls (-ma)
    CPPFLAGS_xlc 	+= -qcpluscmt -M -ma

    #  On 64-bit platforms, add the 64-bit flag
    ifeq "$(_PLATFORM)" "ppu64"
      CPPFLAGS_xlc 	+= -q64
      LDFLAGS_xlc	+= -q64
    endif

  endif

endif

#  Add compiler-specific CPPFLAGS and converted INCDIR
#  Convert the general and compiler-specific INCDIRS into absolute-pathed "-I" options
CPPFLAGS		+=	$(CPPFLAGS_$(COMPILER)) $(foreach i,$(INCDIRS) $(INCDIRS_$(COMPILER)),-I$(abspath $(strip $(i))))

#  Add optimization flags, compiler-specific opt flags, and the compiler-specific CFLAGS
CFLAGS			+=	$(OPTFLAGS) $(OPTFLAGS_$(COMPILER)) $(CFLAGS_$(COMPILER))

#  Add compiler-specific LDFLAGS
LDFLAGS			+=	$(LDFLAGS_$(COMPILER))

#  Add INSTALL_DIR to INSTALL_DIRS in case it was specified
INSTALL_DIRS		+=	$(INSTALL_DIR)

#  Generate a fully-qualified list of install files
INSTALL_TARGETS		=	$(foreach dir,$(INSTALL_DIRS),$(foreach file,$(INSTALL_FILES),$(abspath $(or $(strip $(DESTDIR)),$(topdir))/$(dir)/$(file))))

#  Since the IMPLICIT_SRCS variable has been created above, now cycle through
#  all the PROGRAMS again, and call program_template to create a rule for each
#  one
$(foreach i,$(PROGRAMS),$(eval $(call program_template,$(i))))

#  Set a degault goal
.DEFAULT_GOAL = all

info:
	@echo VPATH_SRCS=$(VPATH_SRCS)
	@echo IMPLICIT_SRCS=$(IMPLICIT_SRCS)
	@echo EXPLICIT_SRCS=$(EXPLICIT_SRCS)
	@echo OBJS=$(OBJS)
	@$(foreach i,$(PROGRAMS),echo $(i); echo "  $(OBJS_$(i))";)
	@echo EMBED32_OBJS=$(EMBED32_OBJS)
	@echo EMBED64_OBJS=$(EMBED64_OBJS)
	@echo LIB_VPATH=$(LIB_VPATH)
	@echo HAVE_ALF_cbea=$(HAVE_ALF_cbea)
	@echo HAVE_ALF=$(HAVE_ALF)
	@echo HAVE_HYBRID_DACS=$(HAVE_HYBRID_DACS)
	@echo HAVE_HYBRID_ALF_x86_64=$(HAVE_HYBRID_ALF_x86_64)
	@echo HAVE_HYBRID_ALF_cbea=$(HAVE_HYBRID_ALF_cbea)
	@echo HAVE_HYBRID_ALF=$(HAVE_HYBRID_ALF)
	@echo HAVE_SPUTIMER=$(HAVE_SPUTIMER)
	@echo HAVE_TOOLS_x86_64=$(HAVE_TOOLS_x86_64)
	@echo HAVE_TOOLS_cbea=$(HAVE_TOOLS_cbea)
	@echo TARGET_CPU=$(TARGET_CPU)
	@echo TARGET_ppc=$(TARGET_ppc)
	@echo TARGET_ppc64=$(TARGET_ppc64)
	@echo BUILD_x86_64=$(BUILD_x86_64)
	@echo BUILD_cbea=$(BUILD_cbea)
	@echo BUILD_HYBRID=$(BUILD_HYBRID)
	@echo BUILD_CELL=$(BUILD_CELL)

#  Mark all the targets which should not short-circuit if they think they are up-to-date
.PHONY: dirs all install uninstall clean 

#  Default rule: 
#
#  1)  Recurse in to directories
#  2)  Invoke any user-supplied targets defined in LOCAL_TARGETS_precompile
#  3)  Build any programs
#  4)  Build any libraries
#  5)  Invoke any user-supplied targets defined in LOCAL_TARGETS_postcompile
all: dirs $(LOCAL_TARGETS_precompile) programs libraries $(LOCAL_TARGETS_postcompile)

#  phony catch-all for any defined DIRS
dirs: $(DIRS)

#  phony catch-all for any defined programs
programs: $(OBJS) $(PROGRAMS)

#  phony catch-all for any defined libraries
libraries: $(LIBRARIES)

#  If there are values for DIRS, define a phony rule for each
ifdef DIRS
.PHONY: $(DIRS)
$(DIRS):
	@if [ -e $@/Makefile ] ; then $(MAKE) -C $@ $(MAKECMDGOALS) ; fi
endif

ifdef EMBED64_OBJS
# Embed into a 32-bit PPU obj
$(EMBED32_OBJS): %-embed.o: %
	$(EMBED_SPU_ppu32) -m32 $(subst -,_,$<) $< $@
endif

ifdef EMBED64_OBJS
# Embed spu obj into a 64-bit PPU obj
$(EMBED64_OBJS): %-embed64.o: %
	$(EMBED_SPU_ppu64) -m64 $(subst -,_,$<) $< $@
endif

$(LIBRARIES_64): $(EMBED64_OBJS)
	$(LD_ppu64) -o $@ -shared $(EMBED64_OBJS) 

$(LIBRARIES_32): $(EMBED32_OBJS) 
	$(LD_ppu32) -o $@ -shared $(EMBED32_OBJS)

install: all $(DIRS)
	@for i in $(INSTALL_TARGETS) ; do \
		if [ ! -d $$(dirname $$i) ] ; then \
			mkdir -p $$(dirname $$i); \
		fi; \
		echo "Installing:  $$i"; \
		$(INSTALL) $$(basename $$i) $$(dirname $$i); \
	done

uninstall: $(DIRS)
	@rm -f $(INSTALL_TARGETS)

clean: $(DIRS) $(LOCAL_TARGETS_clean)
	@rm -f $(CLEAN_TARGETS)

# Dependacy generation rules for .h files (Explicitly required if not using XLC)
ifneq "$(COMPILER)" "xlc"
%.d : %.c
	@set -e; \
	$(CC_$(_PLATFORM)_gnu) -MM $(CPPFLAGS) $< | sed 's,\($*\)\.o[ :]*,\1.o $@ : ,g' > $@
endif

# Include any dependancy files, but don't report errors if not found
$(if $(filter uninstall info clean,$(MAKECMDGOALS)),,$(eval -include $(EXPLICIT_SRCS:.c=.d) $(IMPLICIT_SRCS:.c=.d)))
