# --------------------------------------------------------------  
# (C)Copyright 2007,2008,                                         
# International Business Machines Corporation                     
# All Rights Reserved.                                            
#                                                                 
# Redistribution and use in source and binary forms, with or      
# without modification, are permitted provided that the           
# following conditions are met:                                   
#                                                                 
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer. 
#                                                                 
# - Redistributions in binary form must reproduce the above       
#   copyright notice, this list of conditions and the following   
#   disclaimer in the documentation and/or other materials        
#   provided with the distribution.                               
#                                                                 
# - Neither the name of IBM Corporation nor the names of its      
#   contributors may be used to endorse or promote products       
#   derived from this software without specific prior written     
#   permission.                                                   
#                                                                 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              
# --------------------------------------------------------------  
# PROLOG END TAG zYx                                              

######################################################################
##
##  Basic Configuration
##
######################################################################

#
#  By default, make will attempt to detect what code can be built
#  and will build all that code automatically.
#
#  If this is not desired, these variables can be uncommented and
#  used to disable those checks and explicitly force or disable 
#  a component.  Setting to '1' enables, '0' disables.
#
#BUILD_HYBRID		= 1
#BUILD_CELL		= 1

#
# If no TARGET_CPU is specified, then build the following
# targets when set to "1"
#
ifndef TARGET_CPU
  TARGET_ppc	?= 1
  TARGET_ppc64	?= 1
endif

#
#  The COMPILER variable dicates which compiler suite to use.
#  Valid settings are:  'gnu' and 'xlc'
#
#  Note that gnu is the only compiler for x86_64 binaries
#
COMPILER		?= gnu


#
#  If this variable is set, the ALF and Hybrid ALF includes
#  and libraries will be drawn from this source tree instead
#  of the default system locations.  If set this variable
#  should point to the top of a *built* source tree.
#
#BACKING_TREE		?= 	$(realpath $(topdir)/../../../install)


#
#  These variables control if the built code is linked statically
#  or dynamically against the DaCS and ALF libraries.  Legal values
#  are:  'static' and 'shared'
#
LINKAGE_ALF		?= 	shared
LINKAGE_ALF_HYBRID	?= 	shared
LINKAGE_DACS_HYBRID	?=	shared





######################################################################
##
##  Advanced Configuration
##
######################################################################

#
#  Location of the "install" program
#
INSTALL			=	/usr/bin/install


#
#  Base directories of the SDK for the native architecture
#
SDK_DIR_native		=	/opt/cell/sdk
SDK_PROTOTYPE_DIR_native=	/opt/cell/sdk/prototype

#
#  Base directory of the XLC toolchain
#
XLC_VERSION             =      $(shell ls /opt/ibmcmp/xlc/cbe|sort -rg|head -1)
TOOLCHAIN_XLC           =      "/opt/ibmcmp/xlc/cbe/$(XLC_VERSION)/bin"

#
#  Base directories of the GNU toolchain depending
#  on host architecture
#
ifeq "$(HOST_ARCH)" "ppc64"
  SYSROOT_cbea		=	/
  TOOLCHAIN_cbea	=	/usr
else
  SYSROOT_cbea		=	/opt/cell/sysroot
  TOOLCHAIN_cbea	=	/opt/cell/toolchain

  TOOLCHAIN_x86_64	=	/usr/bin
  SYSROOT_x86_64	=	/
endif


#
#  Base directories of SDK code.
#
#  These will always point to the type of SDK code referred to in their
#  name, regardless of the host platform
#
SDK_DIR_cbea		=	$(SYSROOT_cbea)/opt/cell/sdk
SDK_DIR_x86_64		=	$(SYSROOT_x86_64)/opt/cell/sdk
SDK_PROTOTYPE_DIR_cbea	=	$(SDK_DIR_cbea)/prototype
SDK_PROTOTYPE_DIR_x86_64=	$(SDK_DIR_x86_64)/prototype


INCDIR_SDK_cbea		=	$(SDK_DIR_cbea)/usr/include
INCDIR_SDK_spu		=	$(SDK_DIR_cbea)/usr/spu/include


#
#  GNU tools setup
#
CC_ppu32_gnu		=	$(TOOLCHAIN_cbea)/bin/ppu32-gcc
CC_ppu64_gnu		=	$(TOOLCHAIN_cbea)/bin/ppu-gcc
CC_spu_gnu		=	$(TOOLCHAIN_cbea)/bin/spu-gcc

#
#  XLC tools setup
#
CC_ppu32_xlc		=	$(TOOLCHAIN_XLC)/ppuxlc
CC_ppu64_xlc		=	$(TOOLCHAIN_XLC)/ppuxlc
CC_spu_xlc		=	$(TOOLCHAIN_XLC)/spuxlc

#
#  x86_64 only has a GNU compiler...
#
CC_x86_64_gnu		=	$(TOOLCHAIN_x86_64)/gcc

#
#  Linker setup.  (Both toolchains use the GNU linkers)
#
LD_ppu32		=	$(TOOLCHAIN_cbea)/bin/ppu32-ld
LD_ppu64		=	$(TOOLCHAIN_cbea)/bin/ppu-ld


#
#  Embed SPU setup.  (Both toolchains use the same embedspu)
#
EMBED_SPU_ppu32		=	$(TOOLCHAIN_cbea)/bin/ppu32-embedspu
EMBED_SPU_ppu64		=	$(TOOLCHAIN_cbea)/bin/ppu-embedspu


#
#  Locations of the ALF and Hybrid ALF code 
#
#  (These are the locations of the default installed code)
#
ifndef BACKING_TREE

  #
  #  Hybrid ALF include directories
  #
  INCDIR_ALF_HYBRID_x86_64=	$(SDK_PROTOTYPE_DIR_x86_64)/usr/include
  INCDIR_ALF_HYBRID_spu	=	$(SDK_PROTOTYPE_DIR_cbea)/usr/spu/include 


  #
  #  ALF library directories
  #
  LIBDIR_ALF_ppu32	=	$(SYSROOT_cbea)/usr/lib
  LIBDIR_ALF_ppu64	=	$(SYSROOT_cbea)/usr/lib64
  LIBDIR_ALF_spu	=	$(SYSROOT_cbea)/usr/spu/lib
  
  #
  #  Hybrid ALF library directories
  #
  LIBDIR_ALF_HYBRID_x86_64	=	$(SDK_PROTOTYPE_DIR_x86_64)/usr/lib64
  LIBDIR_ALF_HYBRID_spu		=	$(SDK_PROTOTYPE_DIR_cbea)/usr/spu/lib

  #  
  #  Directories to add to library search path when dynamically linked
  #
  LIBDIR_ALF_RUNTIME_ppu32=	$(SDK_DIR_native)/usr/lib	
  LIBDIR_ALF_RUNTIME_ppu64=	$(SDK_DIR_native)/usr/lib64

#
#  Locations of the ALF and Hybrid ALF code in the backing tree (if defined)
#
else

  #
  #  ALF include directories
  #
  INCDIR_ALF_ppu32	=	$(BACKING_TREE)/noarch/usr/include
  INCDIR_ALF_ppu64	=	$(BACKING_TREE)/noarch/usr/include
  INCDIR_ALF_spu	=	$(BACKING_TREE)/noarch/usr/spu/include

  #  
  #  Hybrid ALF include directories
  #
  INCDIR_ALF_HYBRID_x86_64=	$(BACKING_TREE)/noarch/opt/cell/sdk/prototype/usr/include \
				$(INCDIR_ALF_ppu64)

  INCDIR_ALF_HYBRID_spu	=	$(BACKING_TREE)/noarch/opt/cell/sdk/prototype/usr/spu/include \
				$(INCDIR_ALF_spu)

  #
  #  ALF library directories
  #
  LIBDIR_ALF_ppu32	=	$(BACKING_TREE)/cbea/usr/lib
  LIBDIR_ALF_ppu64	=	$(BACKING_TREE)/cbea/usr/lib64
  LIBDIR_ALF_spu	=	$(BACKING_TREE)/cbea/usr/spu/lib

  #
  #  Hybrid ALF library directories
  #
  LIBDIR_ALF_HYBRID_x86_64=	$(BACKING_TREE)/x86/opt/cell/sdk/prototype/usr/lib64
  LIBDIR_ALF_HYBRID_spu	=	$(BACKING_TREE)/cbea/opt/cell/sdk/prototype/usr/spu/lib

  #
  #  Directories to add to library search path when ALF dynamically linked
  #
  LIBDIR_ALF_RUNTIME_ppu32=	$(LIBDIR_ALF_ppu32)
  LIBDIR_ALF_RUNTIME_ppu64=	$(LIBDIR_ALF_ppu64)

endif

#
#  Directories to add to library search path when Hybrid ALF dynamically linked
#
LIBDIR_ALF_HYBRID_RUNTIME_ppu64		=	$(SDK_PROTOTYPE_DIR_native)/usr/lib64
LIBDIR_ALF_HYBRID_RUNTIME_x86_64	=	$(SDK_PROTOTYPE_DIR_native)/usr/lib64

#
#  DEBUG ALF library directories
#
LIBDIR_ALF_DEBUG_ppu32	=	$(LIBDIR_ALF_ppu32)/alf/debug
LIBDIR_ALF_DEBUG_ppu64	=	$(LIBDIR_ALF_ppu64)/alf/debug
LIBDIR_ALF_DEBUG_spu	=	$(LIBDIR_ALF_spu)/alf/debug

#
#  DEBUG Hybrid ALF library directories
#
LIBDIR_ALF_HYBRID_DEBUG_x86_64	=	$(LIBDIR_ALF_HYBRID_x86_64)/alf/debug
LIBDIR_ALF_HYBRID_DEBUG_spu		=	$(LIBDIR_ALF_HYBRID_spu)/alf/debug

#
#  DEBUG Directories to add to library search path when ALF and Hybrid ALF dynamically linked
#
LIBDIR_ALF_RUNTIME_DEBUG_ppu32	=	$(LIBDIR_ALF_RUNTIME_ppu32)/alf/debug
LIBDIR_ALF_RUNTIME_DEBUG_ppu64	=	$(LIBDIR_ALF_RUNTIME_ppu64)/alf/debug

LIBDIR_ALF_HYBRID_RUNTIME_DEBUG_ppu64	=	$(LIBDIR_ALF_HYBRID_RUNTIME_ppu64)/alf/debug
LIBDIR_ALF_HYBRID_RUNTIME_DEBUG_x86_64	=	$(LIBDIR_ALF_HYBRID_RUNTIME_x86_64)/alf/debug

#
#  Locations of the Hybrid DaCS
#
LIBDIR_DACS_HYBRID_x86_64	=	/usr/lib64

#
#  DEBUG Directories to add to library search path when Hybrid DaCS dynamically linked
#
LIBDIR_DACS_HYBRID_RUNTIME_ppu64	=	/usr/lib64
LIBDIR_DACS_HYBRID_RUNTIME_x86_64	=	/usr/lib64
