/* --------------------------------------------------------------  */
/* (C)Copyright 2008                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <stdio.h>
#include <alf.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

//#define DEBUG

#ifdef DEBUG
#  define debug_print(fmt, arg...) printf(fmt,##arg)
#else
#  define debug_print(fmt, arg...) { }
#endif

#define IMAGE_PATH_BUF_SIZE 1024        // Max length of path to PPU image
char spu_image_path[IMAGE_PATH_BUF_SIZE];       // Used to hold the complete path to SPU image
char library_name[IMAGE_PATH_BUF_SIZE]; // Used to hold the name of spu library
char spu_image_name[] = "alf_hello_world_with_checking_lts_spu";
char task_main[] = "task_main";
int is_SPU_DMA = 1;

#ifdef _ALF_PLATFORM_HYBRID_
char ppu_image_path[IMAGE_PATH_BUF_SIZE];       // Used to hold the complete path to PPU image
#endif

void Usage(char *argv[])
{
  fprintf(stderr, "Usage %s [-h -P]\n", argv[0]);
  fprintf(stderr, "\t-h - help)\n");
  fprintf(stderr, "\t-P - host partitioning (default is accelerator partitioning)\n");
  exit(-1);
}

int main(int argc, char *argv[])
{
  int ret;
  alf_handle_t handle;
  alf_task_desc_handle_t task_desc_handle;
  alf_dataset_handle_t dataset_handle;
  alf_task_handle_t task_handle;
  int ch;

  printf("Hello World beginning!\n");

  is_SPU_DMA = 1;

  while ((ch = getopt(argc, argv, "hP")) != -1) {
    switch (ch) {
    case 'h':
      Usage(argv);
      break;
    case 'P':
      is_SPU_DMA = 0;
      break;
    default:
      Usage(argv);
      break;
    }
  }

#ifdef _ALF_PLATFORM_HYBRID_
  sprintf(library_name, "alf_hello_world_with_checking_lts_hybrid_spu64.so");
#else
  if (sizeof(void *) == 4)
    sprintf(library_name, "alf_hello_world_with_checking_lts_cell_spu.so");
  else
    sprintf(library_name, "alf_hello_world_with_checking_lts_cell_spu64.so");
#endif

  debug_print("Before alf_init\n");
  if ((ret = alf_init(NULL, &handle)) < 0) {
    fprintf(stderr, "Error: alf_init failed, ret=%d\n", ret);
    return 1;
  }

  debug_print("Before alf_num_instances_set\n");
  if ((ret = alf_num_instances_set(handle, 1)) < 0) {
    fprintf(stderr, "Error: alf_num_instances_set failed, ret=%d\n", ret);
    return 1;
  } else if (ret > 0) {
    debug_print("alf_num_instances_set returned number of SPUs=%d\n", ret);
  }
  
  debug_print("Before alf_task_desc_create\n");
  if ((ret = alf_task_desc_create(handle, 0, &task_desc_handle)) < 0) {
    fprintf(stderr, "Error: alf_task_desc_create failed, ret=%d\n", ret);
    return 1;
  } else if (ret > 0) {
    debug_print("alf_task_desc_create returned number of SPUs=%d\n", ret);
  }

  debug_print("Before alf_task_desc_set_int32 of ALF_TASK_DESC_TASK_TYPE=ALF_TASK_TYPE_LIGHTWEIGHT\n");
  if ((ret = alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_TASK_TYPE, ALF_TASK_TYPE_LIGHTWEIGHT)) < 0) {
    fprintf(stderr, "Error: alf_task_desc_set_int32 failed, ret=%d\n", ret);
    return 1;
  }

  debug_print("Before alf_task_desc_set_int32 of ALF_TASK_DESC_MAX_STACK_SIZE=4096\n");
  if ((ret = alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_MAX_STACK_SIZE, 4096)) < 0) {
    fprintf(stderr, "Error: alf_task_desc_set_int32 failed, ret=%d\n", ret);
    return 1;
  }

  debug_print("Before alf_task_desc_set_int32 of ALF_TASK_DESC_WB_PARM_CTX_BUF_SIZE=0\n");
  if ((ret = alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_PARM_CTX_BUF_SIZE, 0)) < 0) {
    fprintf(stderr, "Error: alf_task_desc_set_int32 failed, ret=%d\n", ret);
    return 1;
  }
  
  debug_print("Before alf_task_desc_set_int32 of ALF_TASK_DESC_NUM_DTL_ENTRIES=0\n");
  if ((ret = alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL_ENTRIES, 0)) < 0) {
    fprintf(stderr, "Error: alf_task_desc_set_int32 failed, ret=%d\n", ret);
    return 1;
  }
  
  debug_print("Before alf_task_desc_set_int32 of ALF_TASK_DESC_NUM_DTL=0\n");
  if ((ret = alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL, 0)) < 0) {
    fprintf(stderr, "Error: alf_task_desc_set_int32 failed, ret=%d\n", ret);
    return 1;
  }

  debug_print("Before alf_task_desc_set_int32 of ALF_TASK_DESC_WB_IN_BUF_SIZE=0\n");
  if ((ret = alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_IN_BUF_SIZE, 0)) < 0) {
    fprintf(stderr, "Error: alf_task_desc_set_int32 failed, ret=%d\n", ret);
    return 1;
  }

  debug_print("Before alf_task_desc_set_int32 of ALF_TASK_DESC_WB_OUT_BUF_SIZE=0\n");
  if ((ret = alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_OUT_BUF_SIZE, 0)) < 0) {
    fprintf(stderr, "Error: alf_task_desc_set_int32 failed, ret=%d\n", ret);
    return 1;
  }

  debug_print("Before alf_task_desc_set_int32 of ALF_TASK_DESC_WB_INOUT_BUF_SIZE=0\n");
  if ((ret = alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_INOUT_BUF_SIZE, 0)) < 0) {
    fprintf(stderr, "Error: alf_task_desc_set_int32 failed, ret=%d\n", ret);
    return 1;
  }

  debug_print("Before alf_task_desc_set_int32 of ALF_TASK_DESC_TSK_CTX_SIZE=0\n");
  if ((ret = alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_TSK_CTX_SIZE, 0)) < 0) {
    fprintf(stderr, "Error: alf_task_desc_set_int32 failed, ret=%d\n", ret);
    return 1;
  }

  debug_print("Before alf_task_desc_set_int32 of ALF_TASK_DESC_PARTITION_ON_ACCEL=%d\n", is_SPU_DMA);
  if ((ret =
       alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_PARTITION_ON_ACCEL, (unsigned int) (is_SPU_DMA))) < 0) {
    fprintf(stderr, "Error: alf_task_desc_set_int32 failed, ret=%d\n", ret);
    return 1;
  }

  debug_print("Before alf_task_desc_set_int64 of ALF_TASK_DESC_ACCEL_IMAGE_REF_L=%s\n", spu_image_name);
  if ((ret =
       alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_IMAGE_REF_L,
                               (unsigned long long) spu_image_name)) < 0) {
    fprintf(stderr, "Error: alf_task_desc_set_int64 failed, ret=%d\n", ret);
    return 1;
  }

  debug_print("Before alf_task_desc_set_int64 of ALF_TASK_DESC_ACCEL_LIBRARY_REF_L=%s\n", library_name);
  if ((ret =
       alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_LIBRARY_REF_L,
                               (unsigned long long) library_name)) < 0) {
    fprintf(stderr, "Error: alf_task_desc_set_int64 failed, ret=%d\n", ret);
    return 1;
  }

  debug_print("Before alf_task_desc_set_int64 of ALF_TASK_DESC_ACCEL_LTS_MAIN_REF_L=%s\n", task_main);
  if ((ret =
       alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_LTS_MAIN_REF_L,
                               (unsigned long long) task_main)) < 0) {
    fprintf(stderr, "Error: alf_task_desc_set_int64 failed, ret=%d\n", ret);
    return 1;
  }

  if (is_SPU_DMA) {

    debug_print("Before alf_dataset_create\n");
    if ((ret = alf_dataset_create(handle, &dataset_handle)) < 0) {
      fprintf(stderr, "Error: alf_dataset_create failed, ret=%d\n", ret);
      return 1;
    }

  }

  debug_print("Before alf_task_create\n");
  if ((ret = alf_task_create(task_desc_handle, NULL, 1, 0, 0, &task_handle)) < 0) {
    fprintf(stderr, "Error: alf_task_create failed, ret=%d\n", ret);
    return 1;
  }

  if (is_SPU_DMA) {

    debug_print("Before alf_task_dataset_associate\n");
    if ((ret = alf_task_dataset_associate(task_handle, dataset_handle)) < 0) {
      fprintf(stderr, "Error: alf_task_dataset_associate failed, ret=%d\n", ret);
      return 1;
    }

  }

  debug_print("Before alf_task_desc_destroy\n");
  if ((ret = alf_task_desc_destroy(task_desc_handle)) < 0) {
    fprintf(stderr, "Error: alf_exit alf_task_desc_destroy, ret=%d\n", ret);
    return 1;
  }

  debug_print("Before alf_task_finalize\n");
  if ((ret = alf_task_finalize(task_handle)) < 0) {
    fprintf(stderr, "Error: alf_task_finalize failed, ret=%d\n", ret);
    return 1;
  }

  debug_print("Before alf_task_wait\n");
  if ((ret = alf_task_wait(task_handle, -1)) < 0) {
    fprintf(stderr, "Error: alf_task_wait failed, ret=%d\n", ret);
    return 1;
  } else if (ret > 0) {
    debug_print("alf_task_wait returned number of work blocks=%d\n", ret);
  }
  debug_print("After alf_task_wait\n");

  if (is_SPU_DMA) {

    debug_print("Before alf_dataset_destroy\n");
    if ((ret = alf_dataset_destroy(dataset_handle)) < 0) {
      fprintf(stderr, "Error: alf_dataset_destroy failed, ret=%d\n", ret);
      return 1;
    }

  }

  debug_print("Before alf_exit\n");
  if ((ret = alf_exit(handle, ALF_EXIT_POLICY_FORCE, 0)) < 0) {
    fprintf(stderr, "Error: alf_exit failed, ret=%d\n", ret);
    return 1;
  }

  printf("Hello World completed successfully, exiting!\n");

  return 0;
}
