/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

/***************************************************************
*   PPE side code.This case demonstrates overlay usage
***************************************************************/
//#include <libspe.h>
#include <time.h>
#include <byteswap.h>
#include <string.h>
#include "alf.h"
#include "inverse_matrix_overlay.h"

extern void *malloc_align(size_t size, unsigned int log2_align);
extern void free_align(void *ptr);

#define MAX_SPUS  8
#define BUFFER_SIZE 64

#define PATH_BUF_SIZE 1024      // Max length of path to PPU image
char spu_image_path[PATH_BUF_SIZE];     // Used to hold the complete path to SPU image
char library_name[PATH_BUF_SIZE];       // Used to hold the name of spu library
char spu_image_name[] = "alf_inverse_matrix_ovl_spu";
char kernel_name[] = "comp_kernel";
char input_dtl_name[] = "input_dtl_prepare";

#ifdef _ALF_PLATFORM_HYBRID_
char ppu_image_path[PATH_BUF_SIZE];     // Used to hold the complete path to PPU image
#endif

/*******generate matrix ****************************/
/* This function used to generate 8*8 sub block */
void row_preparation(float *row_p, int size)
{
  int i, j;

  row_p[0] = 1.0f + (rand() % 9) / 10.0f;
  row_p[4 * size + 4] = 1.0f + (rand() % 9) / 10.0f;

  /* initialize 4 element of first row */
  for (i = 1; i < 4; i++) {
    row_p[i] = 1.0f + row_p[i - 1];
    row_p[4 * size + 4 + i] = 1.0f + row_p[4 * size + 4 + i - 1];
  }
  /* initialize 3*4 element of first 4 rows, */
  /* other elements in these 4 rows are zero */
  for (i = 1; i < 4; i++) {
    for (j = 0; j < 4; j++) {
      row_p[j + i * size] = row_p[j] * row_p[j + (i - 1) * size];
    }
  }

  /* initialize 5-8 rows in total 4*8 elements */
  for (i = 5; i < 8; i++) {
    for (j = 4; j < 8; j++) {
      row_p[j + i * size] = row_p[j + 4 * size] * row_p[j + (i - 1) * size];
    }
  }

  for (i = 5; i < 8; i++) {
    for (j = 0; j < 4; j++) {
      row_p[j + i * size] = 1.0f + (rand() % 9) / 10.0f;
    }
  }
}

/* This function prepare special matrix for inversement
 * How the matrix is generated is demonstrated in README */
void matrix_preparation(float *matrix_p, int size)
{
  int i;
  float *row_p;

  for (i = 0; i < size * size; i++) {
    matrix_p[i] = 0;
  }

  for (i = 0; i < size; i = i + 8) {
    row_p = &matrix_p[i * size + i];
    row_preparation(row_p, size);
  }

}

/* This function performs result verification, multiply original 
 * matrix and inverse result to indentify whether it is a unit matrix */
int matrix_verify(float *matrix_p, float *matrix_bk, int size)
{
  int i, j, k;
  float *unit_p = (float *) malloc_align(size * size * sizeof(int), 7);
  float *mul_p = (float *) malloc_align(size * size * sizeof(int), 7);
  int ret = 0;

  /* generate unit matrix */
  for (i = 0; i < size; i++) {
    for (j = 0; j < size; j++) {
      if (i == j)
        unit_p[i * size + j] = 1.0f;
      else
        unit_p[i * size + j] = 0.0f;
      mul_p[i * size + j] = 0.0f;
    }
  }
  /* multiply the original matrix with inverse result matrix */
  for (i = 0; i < size; i++) {
    for (j = 0; j < size; j++) {
      for (k = 0; k < size; k++)
        mul_p[i * size + j] += matrix_bk[i * size + k] * matrix_p[k * size + j];
    }
  }
  /*compare with unit matrix */
  for (i = 0; i < size; i++) {
    for (j = 0; j < size; j++) {
      if (fabs(mul_p[i * size + j] - unit_p[i * size + j]) > 0.005f)
        ret++;
    }
  }

  return ret;
}

int main(int argc __attribute__((unused)), char *argv[] __attribute__((unused)))
{

  int i;
  int buffer_size = BUFFER_SIZE;

  int verify = 0;
  int nodes = 1;
  srand((unsigned) time(NULL));

  void *config_parms = NULL;
  alf_handle_t handle;
  alf_task_desc_handle_t task_desc_handle;
  alf_task_handle_t task_handle;
  alf_wb_handle_t wb_handle;
  alf_dataset_handle_t dataset_handle;

  udata_t my_udata __attribute__ ((aligned(128)));

  float *matrix_p = (float *) malloc_align(buffer_size * buffer_size * sizeof(float), 7);
  float *matrix_bk = (float *) malloc_align(buffer_size * buffer_size * sizeof(float), 7);
  if (matrix_p == NULL || matrix_bk == NULL) {
    printf("matrix malloc failed \n");
    return -1;
  }

  printf("Inverse Matrix Overlay preparing input data\n");

  /* prepare matrix for inverse */
  matrix_preparation(matrix_p, buffer_size);
  memcpy(matrix_bk, matrix_p, buffer_size * buffer_size * sizeof(float));

  my_udata.msize = buffer_size;

  printf("Inverse Matrix Overlay running\n");

  /* begin alf routine */
#ifdef _ALF_PLATFORM_HYBRID_
  sprintf(library_name, "alf_inverse_matrix_ovl_hybrid_spu64.so");
#else
  if (sizeof(void *) == 4)
    sprintf(library_name, "alf_inverse_matrix_ovl_cell_spu.so");
  else
    sprintf(library_name, "alf_inverse_matrix_ovl_cell_spu64.so");
#endif
  alf_init(config_parms, &handle);
  alf_num_instances_set(handle, nodes);

  /* filling out the task info structure */
  alf_task_desc_create(handle, 0, &task_desc_handle);

  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_PARM_CTX_BUF_SIZE, sizeof(udata_t));
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_INOUT_BUF_SIZE, BLOCK_SIZE * BLOCK_SIZE * sizeof(float));
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL_ENTRIES, BLOCK_SIZE);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL, 1);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_PARTITION_ON_ACCEL, 1);

  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_IMAGE_REF_L, (unsigned long long) spu_image_name);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_LIBRARY_REF_L, (unsigned long long) library_name);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_KERNEL_REF_L, (unsigned long long) kernel_name);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_INPUT_DTL_REF_L, (unsigned long long) input_dtl_name);

  alf_dataset_create(handle, &dataset_handle);
  alf_dataset_buffer_add(dataset_handle, matrix_p, buffer_size * buffer_size * sizeof(float), ALF_DATASET_READ_WRITE);

  /* creating task */
  alf_task_create(task_desc_handle, NULL, 0, 0, 1, &task_handle);

  alf_task_dataset_associate(task_handle, dataset_handle);

  /* create wb and add param */
  for (i = 0; i < buffer_size; i = i + 8) {
    alf_data_addr64_t matrix_p_addr64;
    PTR_TO_ADDR64(matrix_p, matrix_p_addr64)
    my_udata.addr_a = matrix_p_addr64 + (i * buffer_size + i) * sizeof(float);

    alf_wb_create(task_handle, ALF_WB_SINGLE, 0, &wb_handle);
    /* Pass matrix pointer and size as parameter */
    alf_wb_parm_add(wb_handle, (void *) &(my_udata.addr_a), 1, ALF_DATA_ADDR64, 3);
    alf_wb_parm_add(wb_handle, (void *) &(my_udata.msize), 1, ALF_DATA_INT32, 2);
    alf_wb_enqueue(wb_handle);
  }
  alf_task_finalize(task_handle);

  /* alf task wait until wb processed */
  alf_task_wait(task_handle, -1);

  alf_task_desc_destroy(task_desc_handle);

  /* end of alf routine */
  alf_exit(handle, ALF_EXIT_POLICY_FORCE, 0);

  printf("Inverse Matrix Overlay running\n");

  /* result verify */
  verify = matrix_verify(matrix_p, matrix_bk, buffer_size);
  if (verify != 0)
    printf("Inverse Matrix Overlay verify failed %d\n", verify);
  else
    printf("Inverse Matrix Overlay verify successful\n");

  free_align(matrix_p);
  free_align(matrix_bk);
  return verify;
}
