/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <alf.h>
#include "matrix_add.h"

#define NUM_ROW 1024
#define NUM_COL 512

#define  H  4
#define  V  NUM_COL

#define PATH_BUF_SIZE 1024  // Max length of path to PPU image
char spu_image_path[PATH_BUF_SIZE];   // Used to hold the complete path to SPU image
char library_name[PATH_BUF_SIZE];     // Used to hold the name of spu library
const char *spu_image_name = "alf_matrix_add_spu";
const char *kernel_name = "comp_kernel";
const char *input_dtl_name = "input_list_prepare";
const char *output_dtl_name = "output_list_prepare";

#ifdef _ALF_PLATFORM_HYBRID_
char ppu_image_path[PATH_BUF_SIZE];    // Used to hold the complete path to PPU image
#endif

float mat_a[NUM_ROW][NUM_COL] __attribute__ ((__aligned__(128)));
float mat_b[NUM_ROW][NUM_COL] __attribute__ ((__aligned__(128)));
float mat_c[NUM_ROW][NUM_COL] __attribute__ ((__aligned__(128)));

int main()
{
  alf_handle_t alf_handle;
  alf_task_desc_handle_t task_desc_handle;
  alf_task_handle_t task_handle;
  alf_wb_handle_t wb_handle;
  add_parms_t parm __attribute__ ((aligned(128)));
  int i, j, rc;
  unsigned int nodes;
  void *config_parms = NULL;
  alf_dataset_handle_t dataset_handle;

  printf("Matrix Addition initializing input.\n");

  /* initialize matrix a b c */
  for (i = 0; i < NUM_ROW; i++) {
    for (j = 0; j < NUM_COL; j++) {
      mat_a[i][j] = (float) (i * NUM_COL + j);
      mat_b[i][j] = (float) ((i * NUM_COL + j) * 2);
      mat_c[i][j] = 0.0f;
    }
  }

  printf("Matrix Addition initialized. Computing results...\n");

  /* begin alf routine */
#ifdef _ALF_PLATFORM_HYBRID_
  /*
  alf_sys_config_t_Hybrid hybrid_config_parms;
  memset(&hybrid_config_parms,0,sizeof(alf_sys_config_t_Hybrid));
  hybrid_config_parms.library_path = getenv("HOME");
  hybrid_config_parms.ppe_image_path = getenv("HOME");
  hybrid_config_parms.num_of_ppes = 1;

  config_parms = &hybrid_config_parms;
  */
  sprintf(library_name, "alf_matrix_add_accel_partition_hybrid_spu64.so");
#else
  /*
  alf_sys_config_t_CBEA cell_config_parms;
  memset(&cell_config_parms,0,sizeof(alf_sys_config_t_CBEA));
  cell_config_parms.library_path = getenv("ALF_SPU_LIB_PATH");

  config_parms = &cell_config_parms;
  */
  if (sizeof(void *) == 4)
    sprintf(library_name, "alf_matrix_add_accel_partition_cell_spu.so");
  else
    sprintf(library_name, "alf_matrix_add_accel_partition_cell_spu64.so");
#endif

  /* Initializing ALF runtime */
  rc = alf_init(config_parms, &alf_handle);
  if (rc < 0) {
    fprintf(stderr, "alf init failed\n");
    return 1;
  } 

  /* query for number of available accelerators on the system */
  rc = alf_query_system_info(alf_handle,ALF_QUERY_NUM_ACCEL,0, &nodes);
  if (rc < 0) {
    fprintf(stderr, "Failed to call alf_query_system_info.\n");
    return 1;
  } else if (nodes <= 0) {
    fprintf(stderr, "Cannot allocate spe to use.\n");
    return 1;
  }

  /* set number of total accelerator instances will be running for this ALF runtime */
  rc = alf_num_instances_set(alf_handle, nodes);
  if (rc < 0) {
    fprintf(stderr, "Cannot init ALF library(%d).\n", rc);
    return rc;
  }
  /* Create and describe a task descriptor */
  alf_task_desc_create(alf_handle, 0,&task_desc_handle);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_PARM_CTX_BUF_SIZE, sizeof(add_parms_t));

  /* we need to reserve enough input space for two input buffers, matrix a and matrix b */
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_IN_BUF_SIZE, 2 * H * V * sizeof(float));  

  /* we need to reserve enough output space for output matrix c */
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_OUT_BUF_SIZE, H * V * sizeof (float));    
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL_ENTRIES, 3);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL, 2);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_MAX_STACK_SIZE, 8192);

  /* set task descriptor to Accelerator partitioning */
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_PARTITION_ON_ACCEL, 1);
  unsigned long long spu_image_name_addr64;
  PTR_TO_ADDR64(spu_image_name, spu_image_name_addr64)
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_IMAGE_REF_L, spu_image_name_addr64);
  unsigned long long library_name_addr64;
  PTR_TO_ADDR64(library_name, library_name_addr64)
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_LIBRARY_REF_L, library_name_addr64);
  unsigned long long kernel_name_addr64;
  PTR_TO_ADDR64(kernel_name, kernel_name_addr64)
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_KERNEL_REF_L, kernel_name_addr64);
  unsigned long long input_dtl_name_addr64;
  PTR_TO_ADDR64(input_dtl_name, input_dtl_name_addr64)
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_INPUT_DTL_REF_L, input_dtl_name_addr64);
  unsigned long long output_dtl_name_addr64;
  PTR_TO_ADDR64(output_dtl_name, output_dtl_name_addr64)
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_OUTPUT_DTL_REF_L, output_dtl_name_addr64);

  /* creating task */
  rc = alf_task_create(task_desc_handle, NULL, 1, 0, 0, &task_handle);
  if (rc < 0) {
    fprintf(stderr, "Cannot create ALF task(%d).\n", rc);
    alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);
    return 1;
  }

  printf("Matrix Addition with data set\n");
  rc = alf_dataset_create(alf_handle, &dataset_handle);
  if (rc < 0) {
    fprintf(stderr, "Call alf_dataset_create error: %d\n", rc);
    return 1;
  }
  rc = alf_dataset_buffer_add(dataset_handle, mat_a, NUM_ROW*NUM_COL*sizeof(float), ALF_DATASET_READ_ONLY);
  if (rc < 0) {
    fprintf(stderr, "Call alf_dataset_buffer_add error: %d\n", rc);
    return 1;
  }
  rc = alf_dataset_buffer_add(dataset_handle, mat_b, NUM_ROW*NUM_COL*sizeof(float), ALF_DATASET_READ_ONLY);
  if (rc < 0) {
    fprintf(stderr, "Call alf_dataset_buffer_add error: %d\n", rc);
    return 1;
  }
  rc = alf_dataset_buffer_add(dataset_handle, mat_c, NUM_ROW*NUM_COL*sizeof(float), ALF_DATASET_WRITE_ONLY);
  if (rc < 0) {
    fprintf(stderr, "Call alf_dataset_buffer_add error: %d\n", rc);
    return 1;
  }
  rc = alf_task_dataset_associate(task_handle, dataset_handle);
  if (rc < 0) {
    fprintf(stderr, "Call alf_task_dataset_associate error: %d\n", rc);
    return 1;
  }


  parm.h = H;
  parm.v = V;
  parm.size = H * V;


  /* creating wb and adding param & io buffer */
  for (i = 0; i < NUM_ROW; i += H) {

    rc = alf_wb_create(task_handle, ALF_WB_SINGLE, 0, &wb_handle);
    if (rc < 0)
    {
      fprintf (stderr, "Call alf_wb_create error: %d\n", rc);
      return 1;

    }

    PTR_TO_ADDR64(&mat_a[i][0], parm.mat_a);
    PTR_TO_ADDR64(&mat_b[i][0], parm.mat_b);
    PTR_TO_ADDR64(&mat_c[i][0], parm.mat_c);

    alf_wb_parm_add(wb_handle, (void *) (&parm.h), 1, ALF_DATA_INT32, 0);
    alf_wb_parm_add(wb_handle, (void *) (&parm.v), 1, ALF_DATA_INT32, 0);
    alf_wb_parm_add(wb_handle, (void *) (&parm.mat_a), 1, ALF_DATA_INT64, 0);
    alf_wb_parm_add(wb_handle, (void *) (&parm.mat_b), 1, ALF_DATA_INT64, 0);
    alf_wb_parm_add(wb_handle, (void *) (&parm.mat_c), 1, ALF_DATA_INT64, 0);
    alf_wb_parm_add(wb_handle, (void *) (&parm.size), 1, ALF_DATA_INT32, 0);
    
    rc = alf_wb_enqueue(wb_handle);
    if (rc < 0) {
      fprintf(stderr, "Call alf_wb_enqueue error: %d\n", rc);
      return 1;
    }
  }

  alf_task_finalize(task_handle);

  /* alf task wait until wb processed */
  alf_task_wait(task_handle, -1);
  alf_task_destroy(task_handle);
  /* end of alf routine */

  alf_exit(alf_handle, ALF_EXIT_POLICY_WAIT, -1);
  
  printf("Matrix Addition computed. Verifying results...\n");

  /* verifying */
  for (i = 0; i < NUM_ROW; i++) {
    for (j = 0; j < NUM_COL; j++) {
      if (mat_c[i][j] != (float) ((i * NUM_COL + j) * 3)) {
        printf("Matrix Addition failed verification. c[%d][%d] != %f\n",i,j,(float) ((i * NUM_COL + j) * 3));
        return -1;
      }
    }
  }

  printf("Matrix Addition successfully verified.\n");

  return 0;
}
