/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <alf.h>
#include "matrix_add.h"

#define NUM_ROW 1024
#define NUM_COL 512

#define  H  4
#define  V  NUM_COL

#define PATH_BUF_SIZE 1024  // Max length of path to PPU image
char spu_image_path[PATH_BUF_SIZE];   // Used to hold the complete path to SPU image
char library_name[PATH_BUF_SIZE];     // Used to hold the name of spu library
const char *spu_image_name = "alf_matrix_add_spu";
const char *kernel_name = "comp_kernel";
const char *input_dtl_name = "input_list_prepare";
const char *output_dtl_name = "output_list_prepare";

#ifdef _ALF_PLATFORM_HYBRID_
char ppu_image_path[PATH_BUF_SIZE];    // Used to hold the complete path to PPU image
#endif

float mat_a[NUM_ROW][NUM_COL] __attribute__ ((__aligned__(128)));
float mat_b[NUM_ROW][NUM_COL] __attribute__ ((__aligned__(128)));

int main(int argc, char **argv)
{
  alf_handle_t alf_handle;
  alf_task_desc_handle_t task_desc_handle;
  alf_task_handle_t task_handle;
  alf_wb_handle_t wb_handle;
  add_parms_t parm __attribute__ ((aligned(128)));
  int i, j, rc;
  unsigned int nodes;
  int is_spu_dma = 0;
  void *config_parms = NULL;

  if (argc == 2) {
    is_spu_dma = atoi(argv[1]);
    if (is_spu_dma < 0)
      is_spu_dma = 0;
  }
  
  printf("is_spu_dma: %d\n", is_spu_dma);
  
#ifdef _ALF_PLATFORM_HYBRID_
  if (is_spu_dma == 1) {
     fprintf(stderr, "is_spu_dma value of 1 not supported in hybrid environment\n");
     return 1;
  }
#endif

  printf("Matrix Addition initializing input.\n");

  /* initialize matrix a b c */
  for (i = 0; i < NUM_ROW; i++) {
    for (j = 0; j < NUM_COL; j++) {
      mat_a[i][j] = (float) (i * NUM_COL + j);
      mat_b[i][j] = (float) ((i * NUM_COL + j) * 2);
    }
  }

  printf("Matrix Addition initialized. Computing results...\n");

  /* begin alf routine */
#ifdef _ALF_PLATFORM_HYBRID_
 sprintf(library_name, "alf_matrix_add_overlapped_io_hybrid_spu64.so");
#else
 if (sizeof(void *) == 4)
    sprintf(library_name, "alf_matrix_add_overlapped_io_cell_spu.so");
  else
    sprintf(library_name, "alf_matrix_add_overlapped_io_cell_spu64.so");
#endif

  rc = alf_init(config_parms, &alf_handle);
  if (rc < 0) {
    fprintf(stderr, "alf init failed\n");
    return 1;
  } 

  rc = alf_query_system_info(alf_handle,ALF_QUERY_NUM_ACCEL,0, &nodes);
  if (rc < 0) {
    fprintf(stderr, "Failed to call alf_query_system_info.\n");
    return 1;
  } else if (nodes <= 0) {
    fprintf(stderr, "Cannot allocate spe to use.\n");
    return 1;
  } else if (nodes > 2) {
    /* Only use 2 SPEs in this demo */
    nodes = 2;
  }

  rc = alf_num_instances_set(alf_handle, nodes);
  if (rc < 0) {
    fprintf(stderr, "Cannot init ALF library(%d).\n", rc);
    return rc;
  }

  alf_task_desc_create(alf_handle, 0,&task_desc_handle);

  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_PARM_CTX_BUF_SIZE, sizeof(add_parms_t));
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_IN_BUF_SIZE, H * V * sizeof(float)); // 32k
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_OUT_BUF_SIZE, 0);    // 0k
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_INOUT_BUF_SIZE,  H * V * sizeof(float));
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL_ENTRIES, 2);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL, 2);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_TSK_CTX_SIZE, 0);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_MAX_STACK_SIZE, 8192);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_PARTITION_ON_ACCEL, is_spu_dma);
  unsigned long long spu_image_name_addr64;
  PTR_TO_ADDR64(spu_image_name, spu_image_name_addr64)
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_IMAGE_REF_L, spu_image_name_addr64);
  unsigned long long library_name_addr64;
  PTR_TO_ADDR64(library_name, library_name_addr64)
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_LIBRARY_REF_L, library_name_addr64);
  unsigned long long kernel_name_addr64;
  PTR_TO_ADDR64(kernel_name, kernel_name_addr64)
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_KERNEL_REF_L, kernel_name_addr64);
  unsigned long long input_dtl_name_addr64;
  PTR_TO_ADDR64(input_dtl_name, input_dtl_name_addr64)
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_INPUT_DTL_REF_L, input_dtl_name_addr64);

  /* creating task */
  rc = alf_task_create(task_desc_handle, NULL, 0, 0, 0, &task_handle);
  if (rc < 0) {
    fprintf(stderr, "Cannot create ALF task(%d).\n", rc);
    alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);
    return 1;
  }
  
  parm.h = H;
  parm.v = V;
  parm.size = H * V;


  /* creating wb and adding param & io buffer */
  for (i = 0; i < NUM_ROW; i += H) {

    alf_wb_create(task_handle, ALF_WB_SINGLE, 0, &wb_handle);
    
    if (!is_spu_dma) {
      alf_wb_dtl_begin(wb_handle, ALF_BUF_IN, 0);
      alf_wb_dtl_entry_add(wb_handle, &mat_a[i][0], H * V, ALF_DATA_FLOAT);
      alf_wb_dtl_end(wb_handle);

      alf_wb_dtl_begin(wb_handle, ALF_BUF_OVL_INOUT, 0);
      alf_wb_dtl_entry_add(wb_handle, &mat_b[i][0], H * V, ALF_DATA_FLOAT);
      alf_wb_dtl_end(wb_handle);

      parm.in_data = 0;
      parm.out_data = 0;
    } else {
      PTR_TO_ADDR64(&mat_a[i][0],parm.in_data);
      PTR_TO_ADDR64(&mat_b[i][0],parm.out_data);
    }

    alf_wb_parm_add(wb_handle, (void *) (&parm.h), 1, ALF_DATA_INT32, 0);
    alf_wb_parm_add(wb_handle, (void *) (&parm.v), 1, ALF_DATA_INT32, 0);
    alf_wb_parm_add(wb_handle, (void *) (&parm.in_data), 1, ALF_DATA_INT64, 0);
    alf_wb_parm_add(wb_handle, (void *) (&parm.out_data), 1, ALF_DATA_INT64, 0);
    alf_wb_parm_add(wb_handle, (void *) (&parm.size), 1, ALF_DATA_INT32, 0);
    
    rc = alf_wb_enqueue(wb_handle);
    if (rc < 0) {
      fprintf(stderr, "Call alf_wb_enqueue error: %d\n", rc);
      return 1;
    }
  }

  alf_task_finalize(task_handle);

  /* alf task wait until wb processed */
  alf_task_wait(task_handle, -1);
  alf_task_destroy(task_handle);
  /* end of alf routine */

  alf_exit(alf_handle, ALF_EXIT_POLICY_WAIT, -1);
  
  printf("Matrix Addition computed. Verifying results...\n");

  /* verifying */
  for (i = 0; i < NUM_ROW; i++) {
    for (j = 0; j < NUM_COL; j++) {
      if (mat_b[i][j] != (float) ((i * NUM_COL + j) * 3)) {
        printf("Matrix Addition failed verification. b[%d][%d] != %f\n",i,j,(float) ((i * NUM_COL + j) * 3));
        return -1;
      }
    }
  }

  printf("Matrix Addition successfully verified.\n");

  return 0;
}
