/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <alf_accel.h>
#include "../matrix_add.h"

int comp_kernel(void *p_task_context __attribute__ ((unused)),
                          void *p_parm_ctx_buffer,
                          void *p_input_buffer,
                          void *p_output_buffer __attribute__ ((unused)),
                          void *p_inout_buffer,
                          unsigned int current_count __attribute__ ((unused)),
                          unsigned int total_count __attribute__ ((unused)))
{
  unsigned int i, cnt;
  static int count;
  float *sa, *sb, *sc;
  void *p_parm = p_parm_ctx_buffer;
  unsigned int h = *((unsigned int*)p_parm);
  p_parm += sizeof(h);
  unsigned int v = *((unsigned int*)p_parm);

  cnt = h * v;
  sa = (float *) p_input_buffer;
  sb = (float *) p_inout_buffer;
  sc = sb;
  for (i = 0; i < cnt; i++) {
    sc[i] = sa[i] + sb[i];
    count++;
  }
  return 0;
}

int input_list_prepare(void *p_task_context __attribute__ ((unused)),
                                 void *p_parm_ctx_buffer,
                                 void *p_dtl,
                                 unsigned int current_count __attribute__ ((unused)),
                                 unsigned int total_count __attribute__ ((unused)))
{
  alf_data_addr64_t ea;
  void *p_parm = p_parm_ctx_buffer;
  p_parm += sizeof(unsigned int)+sizeof(unsigned int);//skip h and v
  unsigned long long in_data = *((unsigned long long*)p_parm);
  p_parm += sizeof(in_data);
  unsigned long long out_data = *((unsigned long long*)p_parm);
  p_parm += sizeof(out_data);
  unsigned int size = *((unsigned int*)p_parm);

  ALF_ACCEL_DTL_BEGIN(p_dtl, ALF_BUF_IN, 0);
  ea = (unsigned int)in_data;
  ALF_ACCEL_DTL_ENTRY_ADD(p_dtl, size, ALF_DATA_FLOAT, ea);
  ALF_ACCEL_DTL_END(p_dtl);

  ALF_ACCEL_DTL_BEGIN(p_dtl, ALF_BUF_OVL_INOUT, 0);
  ea = (unsigned int)out_data;
  ALF_ACCEL_DTL_ENTRY_ADD(p_dtl, size, ALF_DATA_FLOAT, ea);
  ALF_ACCEL_DTL_END(p_dtl);

  return 0;
}

ALF_ACCEL_EXPORT_API_LIST_BEGIN
   ALF_ACCEL_EXPORT_API("", comp_kernel);
   ALF_ACCEL_EXPORT_API("", input_list_prepare);
ALF_ACCEL_EXPORT_API_LIST_END

