/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <alf.h>
#include "matrix_trans.h"

#define NUM_ROW		1024
#define NUM_COL		512

#define  H  128
#define  V  128

#define MY_ALIGN(_my_var_def_, _my_al_)  _my_var_def_  \
__attribute__((__aligned__(_my_al_)))

MY_ALIGN(float mat_a[NUM_ROW][NUM_COL], 128);
MY_ALIGN(float mat_c[NUM_COL][NUM_ROW], 128);

#define PATH_BUF_SIZE 1024      // Max length of path to PPU image
char spu_image_path[PATH_BUF_SIZE];     // Used to hold the complete path to SPU image
char library_name[PATH_BUF_SIZE];       // Used to hold the name of spu library
char spu_image_name[] = "alf_matrix_transpose_spu";
char kernel_name[] = "comp_kernel";

#ifdef _ALF_PLATFORM_HYBRID_
char ppu_image_path[PATH_BUF_SIZE];     // Used to hold the complete path to PPU image
#endif

/* This function perform matrix a c initialization */
void init_mat()
{
  int i, j;

  for (i = 0; i < NUM_COL; i++) {
    for (j = 0; j < NUM_ROW; j++) {
      mat_a[j][i] = j * NUM_COL + i;
      mat_c[i][j] = 0.0f;
    }
  }
}



int main(void)
{
  alf_handle_t alf_handle;
  alf_task_desc_handle_t task_desc_handle;
  alf_task_handle_t task_handle;
  alf_wb_handle_t wb_handle;
  trans_parms_t parm;
  unsigned int nodes;
  void *config_parms = NULL;
  int X, Y, i, j, rc;

  printf("Tranposing matrix preparing input matrix\n");

  /* prepare data */
  init_mat();

  printf("Tranposing matrix running\n");

  /* begin alf routine */
#ifdef _ALF_PLATFORM_HYBRID_
  sprintf(library_name, "alf_matrix_transpose_step1a_hybrid_spu64.so");
#else
  if (sizeof(void *) == 4)
    sprintf(library_name, "alf_matrix_transpose_step1a_cell_spu.so");
  else
    sprintf(library_name, "alf_matrix_transpose_step1a_cell_spu64.so");
#endif

  rc = alf_init(config_parms, &alf_handle);
  if (rc < 0) {
    fprintf(stderr, "alf init failed\n");
    return 1;
  }

  rc = alf_query_system_info(alf_handle, ALF_QUERY_NUM_ACCEL, 0, &nodes);
  if (rc < 0) {
    fprintf(stderr, "Failed to call alf_query_system_info.\n");
    return 1;
  } else if (nodes <= 0) {
    fprintf(stderr, "Cannot allocate spe to use.\n");
    return 1;
  }

  rc = alf_num_instances_set(alf_handle, 1);
  if (rc < 0) {
    alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, -1);
    fprintf(stderr, "Cannot init ALF library(%d).\n", rc);
    return rc;
  }

  alf_task_desc_create(alf_handle, ALF_ACCEL_TYPE_SPE, &task_desc_handle);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_TSK_CTX_SIZE, 0);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_PARM_CTX_BUF_SIZE, sizeof(trans_parms_t));
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_IN_BUF_SIZE, H * V * sizeof(float));       // 64k
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_WB_OUT_BUF_SIZE, H * V * sizeof(float));      // 64k
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_PARTITION_ON_ACCEL, 0);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL_ENTRIES, H + V);
  alf_task_desc_set_int32(task_desc_handle, ALF_TASK_DESC_NUM_DTL, 2);

  /* Set accelerator image name */
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_IMAGE_REF_L, (unsigned long long) spu_image_name);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_LIBRARY_REF_L, (unsigned long long) library_name);
  alf_task_desc_set_int64(task_desc_handle, ALF_TASK_DESC_ACCEL_KERNEL_REF_L, (unsigned long long) kernel_name);

  /* creating task */
  rc = alf_task_create(task_desc_handle, NULL, 1, 0, 0, &task_handle);
  if (rc < 0) {
    fprintf(stderr, "Cannot create ALF task(%d).\n", rc);
    alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, -1);
    return 1;
  };

  parm.h = H;
  parm.v = V;

  /* creating wb and adding param & io buffer */
  for (X = 0; X < NUM_ROW; X += H) {
    for (Y = 0; Y < NUM_COL; Y += V) {
      alf_wb_create(task_handle, ALF_WB_SINGLE, 0, &wb_handle);
      /* Pass H and V to spu */
      alf_wb_parm_add(wb_handle, &(parm.h), 1, ALF_DATA_INT32, 0);
      alf_wb_parm_add(wb_handle, &(parm.v), 1, ALF_DATA_INT32, 0);

      alf_wb_dtl_begin(wb_handle, ALF_BUF_IN, 0);
      /* Add V element of mat_a as input */
      for (i = 0; i < H; i++) {
        alf_wb_dtl_entry_add(wb_handle, &mat_a[X + i][Y], V, ALF_DATA_FLOAT);
      }
      alf_wb_dtl_end(wb_handle);

      alf_wb_dtl_begin(wb_handle, ALF_BUF_OUT, 0);
      /* Add V element of mat_c as output */
      for (i = 0; i < V; i++) {
        alf_wb_dtl_entry_add(wb_handle, &mat_c[Y + i][X], H, ALF_DATA_FLOAT);
      }
      alf_wb_dtl_end(wb_handle);

      alf_wb_enqueue(wb_handle);
    }
  }
  alf_task_finalize(task_handle);

  /* alf task wait until wb processed */
  alf_task_wait(task_handle, -1);

  /* end of alf routine */
  alf_exit(alf_handle, ALF_EXIT_POLICY_WAIT, -1);

  printf("Tranposing matrix verifying...\n");

  //verifying
  for (i = 0; i < NUM_COL; i++) {
    for (j = 0; j < NUM_ROW; j++) {
      if (mat_c[i][j] != mat_a[j][i]) {
        fprintf(stderr, "ERROR from matrix transpose - Exit\n");
        fflush(stderr);
        exit(1);
      }
    }
  }
  
  printf("Tranposing matrix successfully verified!\n");

  return 0;
}
