/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <alf_accel.h> 
#include "../matrix_trans.h"


int comp_kernel(void *p_task_context __attribute__ ((unused)),
                void *p_parm_ctx_buffer,
                void *p_input_buffer,
                void *p_output_buffer,
                void *p_inout_buffer __attribute__ ((unused)),
                unsigned int current_count __attribute__ ((unused)),
                unsigned int total_count __attribute__ ((unused)))
{
    unsigned int i, j;
    float *sa, *sc;
    trans_parms_t *p_parm = (trans_parms_t *)p_parm_ctx_buffer;

    sa = (float *) p_input_buffer;
    sc = (float *) p_output_buffer;

    for(i=0; i< p_parm->h; i++)
      for(j=0; j< p_parm->v; j++)
          *(sc+j*p_parm->h + i)      /* sc[j][i] */
          = *(sa+i*p_parm->v +j);    /* sa[i][j] */

    return 0;
}

int input_list_prepare(void *p_task_context __attribute__ ((unused)),
                       void *p_parm_ctx_buffer,
                       void *p_dtl,
                       unsigned int current_count __attribute__ ((unused)),
                       unsigned int total_count __attribute__ ((unused)))
{
     trans_parms_t *p_parm = (trans_parms_t *)p_parm_ctx_buffer;
     unsigned long long pA;
     unsigned int i;
     alf_data_addr64_t  ea;

     pA = p_parm->p_mat_a + 
           sizeof(float) * (p_parm->DIMY * p_parm->X + p_parm->Y); // mat_a[X][Y]

     ALF_ACCEL_DTL_BEGIN(p_dtl, ALF_BUF_IN, 0);

     for(i=0; i<p_parm->h; i++)
     {
        ea = pA + p_parm->DIMY * i * sizeof(float); // mat_a[X+i][Y]
        ALF_ACCEL_DTL_ENTRY_ADD(p_dtl, p_parm->v, ALF_DATA_FLOAT, ea);
     }

     ALF_ACCEL_DTL_END(p_dtl);
     return 0;
}

int output_list_prepare(void *p_task_context __attribute__ ((unused)),
                        void *p_parm_ctx_buffer,
                        void *p_dtl,
                        unsigned int current_count __attribute__ ((unused)),
                        unsigned int total_count __attribute__ ((unused)))
{
     trans_parms_t *p_parm = (trans_parms_t *)p_parm_ctx_buffer;
     unsigned long long pC;
     unsigned int j;
     alf_data_addr64_t  ea;

     pC = p_parm->p_mat_c + 
           sizeof(float) * (p_parm->DIMX * p_parm->Y + p_parm->X); // mat_c[Y][X]

     ALF_ACCEL_DTL_BEGIN(p_dtl, ALF_BUF_OUT, 0);

     for(j=0; j<p_parm->v; j++)
     {
        ea = pC + p_parm->DIMX * j * sizeof(float); // mat_c[Y+j][X]
        ALF_ACCEL_DTL_ENTRY_ADD(p_dtl, p_parm->h, ALF_DATA_FLOAT, ea);
     }

     ALF_ACCEL_DTL_END(p_dtl);
     return 0;
}

ALF_ACCEL_EXPORT_API_LIST_BEGIN
   ALF_ACCEL_EXPORT_API("", comp_kernel);
   ALF_ACCEL_EXPORT_API("", input_list_prepare);
   ALF_ACCEL_EXPORT_API("", output_list_prepare);
ALF_ACCEL_EXPORT_API_LIST_END
