/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <stdlib.h>

#include "my_rng.h"

// ---------------------------------------------
// this is a simple Lagged Fibonacci PRNG
// ---------------------------------------------

// the RNG initializer
void init_lfrng(my_context_rng_t *p_ctx, unsigned int seed)
{
	unsigned int i;

	// use system RNG to init it
	srand(seed);

	for(i=0; i<RNG_K; i++)
		p_ctx->s[i] = rand();
	p_ctx->n = 0;

	return ;
}

// the RNG 
unsigned int lfrng(my_context_rng_t *p_ctx)
{
    unsigned int v;
 
	// Sn = ( Sn-j * Sn-k ) mod 2^32,  0 < j < k
	// because we have size k history, so n-k == 0
	// we do not need mod 2^32 as our word length is 32 bits
	p_ctx->s[p_ctx->n] = p_ctx->s[(p_ctx->n + RNG_K - RNG_J)%RNG_K] + p_ctx->s[p_ctx->n];  
	v = p_ctx->s[p_ctx->n];
	p_ctx->n ++;
	if(p_ctx->n >= RNG_K)
		p_ctx->n = 0;
    return v;
}


