/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

/***************************************************************
*   Control  side code
***************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string.h>

#include <alf.h>

#include "pipe_line.h"
#include "my_rng.h"
#include "sim.h"

/*
  * common definition section for a typical ALF application for SDK 3.0 API
  */

const char *app_name = "Pipe Line Simulation";
#define ALF_APP_ID "pipe_line"

#ifdef _ALF_PLATFORM_HYBRID_
    const char *app_library_name_1 = "alf_" ALF_APP_ID "_hybrid_spu64.so";
    const char *app_library_name_2 = "alf_" ALF_APP_ID "_hybrid_spu64.so";
#elif defined(_ALF_PLATFORM_CELL_)
    #ifdef _ALF_64B_
        const char *app_library_name_1 = "alf_" ALF_APP_ID "_cell_spu64.so";
        const char *app_library_name_2 = "alf_" ALF_APP_ID "_cell_spu64.so";
    #else
        const char *app_library_name_1 = "alf_" ALF_APP_ID "_cell_spu.so";
        const char *app_library_name_2 = "alf_" ALF_APP_ID "_cell_spu.so";
    #endif
#else
   #error "undefined platform"
#endif

const char *app_image_name_1 = "alf_" ALF_APP_ID "_spu";
const char *app_kernel_name_1 = "rng_kernel";
const char *app_ctx_merge_name_1 = "context_merge";
const char *app_ctx_setup_name_1 = NULL;
const char *app_prep_in_name_1 = NULL;
const char *app_prep_out_name_1 = "rng_prep_out";


const char *app_image_name_2 = "alf_" ALF_APP_ID "_spu";
const char *app_kernel_name_2 = "sim_kernel";
const char *app_ctx_merge_name_2 = "context_merge";
const char *app_ctx_setup_name_2 = NULL;
const char *app_prep_in_name_2 = "sim_prep_in";
const char *app_prep_out_name_2 = NULL;


/*
  * a small init routine to offload the hassle
  */

int app_init_alf(unsigned int num_accel, alf_handle_t *half)
{
    int rc;
    unsigned int nodes;
    void * config_parms = NULL;


      *half = ALF_NULL_HANDLE;
  
      rc = alf_init(config_parms, half);
      if (rc < 0) 
      {
          fprintf(stderr, "**ERR: alf_init failed (%d)\n", rc);
          goto init_err;
      } 
    
      rc = alf_query_system_info(*half,ALF_QUERY_NUM_ACCEL,0, &nodes);
      if (rc < 0) 
      {
          fprintf(stderr, "**ERR: alf_query_system_info failed (%d)\n", rc);
          goto init_err;
      } 
      else if (nodes == 0 || nodes < num_accel) 
      {
          fprintf(stderr, "**ERR: do not have enough accelerators. (have %d, req %d)\n", nodes, num_accel);
          goto init_err;
      }
      
      if(num_accel == 0)
         num_accel = nodes;
      
      rc = alf_num_instances_set(*half, num_accel);
      if(rc < 0)
      {
          fprintf(stderr, "**ERR: alf_num_instances_set(%d) failed (%d)\n", num_accel, rc);
          goto init_err;
      }
      
      return 0;

init_err:

    if(*half)
        alf_exit(*half, ALF_EXIT_POLICY_FORCE, 0);

    *half = ALF_NULL_HANDLE;
    
    return -1;
}




/*
  *  app code begin here 
  */


typedef union _stage1_ctx_t_
{
    my_context_rng_t rng;
    char dummy[(sizeof(my_context_rng_t)+16)&0xfff0];   // a trick to define aligned data structure  
} stage1_ctx_t;

typedef union _stage2_ctx_t_
{
    my_context_sim_t sim;
    char dummy[(sizeof(my_context_sim_t)+16)&0xfff0];   // a trick to define aligned data structure  
} stage2_ctx_t;


stage1_ctx_t stage1_ctx  __attribute__ ((aligned (128)));  // the rng context
stage2_ctx_t stage2_ctx  __attribute__ ((aligned (128)));  // the sim context
alf_data_int32_t stage_buffer[2][BLOCK_SIZE*WB_PER_TASK]  __attribute__ ((aligned (128)));  // the stage  buffer

my_context_rng_t rng_ctx_verify;



/* This function performs data preparation for the application */
void prepare_data()
{
  init_lfrng(&stage1_ctx.rng, 0);
  init_sim(&stage2_ctx.sim);
  memcpy(&rng_ctx_verify, &stage1_ctx.rng, sizeof(rng_ctx_verify)); 
}

int verify_results()
{
    my_context_sim_t sim;
    unsigned int i;
    
    init_sim(&sim);
    for(i=0; i<TOTAL_SIM_TIME/2; i++)
    {
       unsigned int rx, ry;
       rx = lfrng(&rng_ctx_verify);
       ry = lfrng(&rng_ctx_verify);
       sim_one_step(&sim, rx, ry);
    }
    
	printf("Expected %d  left %d,  right %d,  lower %d,  upper %d\n", sim.total_sims, sim.left, sim.right, sim.lower, sim.upper);
  
	printf("Returned %d  left %d,  right %d,  lower %d,  upper %d\n", stage2_ctx.sim.total_sims, stage2_ctx.sim.left, stage2_ctx.sim.right, stage2_ctx.sim.lower, stage2_ctx.sim.upper);

    if(memcmp(&sim, &stage2_ctx.sim, sizeof(my_context_sim_t)))
    {
        fprintf(stderr, "**ERR: verification failed\n");
        return -1;
    }

    return 0;
}
    


int main(void)
{
  // basic variables
  alf_handle_t alf_handle;
  int rc;

  // app specific variables 
  alf_task_desc_handle_t stage1_desc_handle, stage2_desc_handle;
  alf_task_handle_t old_tsk_handles[2][2];
  alf_task_handle_t tsk1, tsk2;
  alf_wb_handle_t wb_handle;
  my_wb_parms_t  wb_parm;
  alf_dataset_handle_t dataset_handle[2];
  unsigned int i;


  printf("%s is initializing ALF runtime.\n", app_name); fflush(stdout);
  // basic routines  
  if(app_init_alf(0, &alf_handle) < 0)
  {
      fprintf(stderr, "**ERR: failed to init ALF runtime\n");
      return -1;
  }


  // now comes to the task description and task creation
  rc = alf_task_desc_create(alf_handle, 0, &stage1_desc_handle);
  if(rc < 0)
  {
      fprintf(stderr, "**ERR: alf_task_desc_create (%d)\n", rc);
      exit(-1);
  }

  
  // I'm too tired to check return code of all of these :-)
  // these are common stuffs for most tasks
  alf_task_desc_set_int64(stage1_desc_handle, ALF_TASK_DESC_ACCEL_LIBRARY_REF_L, (unsigned long)app_library_name_1);
  alf_task_desc_set_int64(stage1_desc_handle, ALF_TASK_DESC_ACCEL_IMAGE_REF_L, (unsigned long)app_image_name_1);
  alf_task_desc_set_int64(stage1_desc_handle, ALF_TASK_DESC_ACCEL_KERNEL_REF_L, (unsigned long)app_kernel_name_1);
  alf_task_desc_set_int64(stage1_desc_handle, ALF_TASK_DESC_ACCEL_CTX_SETUP_REF_L, (unsigned long)app_ctx_setup_name_1);
  alf_task_desc_set_int64(stage1_desc_handle, ALF_TASK_DESC_ACCEL_CTX_MERGE_REF_L, (unsigned long)app_ctx_merge_name_1);
  alf_task_desc_set_int64(stage1_desc_handle, ALF_TASK_DESC_ACCEL_INPUT_DTL_REF_L, (unsigned long)app_prep_in_name_1);
  alf_task_desc_set_int64(stage1_desc_handle, ALF_TASK_DESC_ACCEL_OUTPUT_DTL_REF_L, (unsigned long)app_prep_out_name_1);

  // these are app specific
  alf_task_desc_set_int32(stage1_desc_handle, ALF_TASK_DESC_TSK_CTX_SIZE, sizeof(stage1_ctx_t));
  alf_task_desc_set_int32(stage1_desc_handle, ALF_TASK_DESC_WB_PARM_CTX_BUF_SIZE, sizeof(my_wb_parms_t));
  alf_task_desc_set_int32(stage1_desc_handle, ALF_TASK_DESC_WB_IN_BUF_SIZE, 0);
  // stage1 is out only
  alf_task_desc_set_int32(stage1_desc_handle, ALF_TASK_DESC_WB_OUT_BUF_SIZE, BLOCK_SIZE*sizeof(alf_data_int32_t));
  alf_task_desc_set_int32(stage1_desc_handle, ALF_TASK_DESC_WB_INOUT_BUF_SIZE, 0);
  alf_task_desc_set_int32(stage1_desc_handle, ALF_TASK_DESC_NUM_DTL_ENTRIES, BLOCK_SIZE/4096);
  alf_task_desc_set_int32(stage1_desc_handle, ALF_TASK_DESC_NUM_DTL, 1);
  alf_task_desc_set_int32(stage1_desc_handle, ALF_TASK_DESC_MAX_STACK_SIZE, 1024);
  alf_task_desc_set_int32(stage1_desc_handle, ALF_TASK_DESC_PARTITION_ON_ACCEL, 1); // for host side performance
  // the task context  
  alf_task_desc_ctx_entry_add(stage1_desc_handle, ALF_DATA_INT32, sizeof(my_context_rng_t)/sizeof(alf_data_int32_t));



  // stage 2
  rc = alf_task_desc_create(alf_handle, 0, &stage2_desc_handle);
  if(rc < 0)
  {
      fprintf(stderr, "**ERR: alf_task_desc_create (%d)\n", rc);
      exit(-1);
  }

  
  // I'm too tired to check return code of all of these :-)
  // these are common stuffs for most tasks
  alf_task_desc_set_int64(stage2_desc_handle, ALF_TASK_DESC_ACCEL_LIBRARY_REF_L, (unsigned long)app_library_name_2);
  alf_task_desc_set_int64(stage2_desc_handle, ALF_TASK_DESC_ACCEL_IMAGE_REF_L, (unsigned long)app_image_name_2);
  alf_task_desc_set_int64(stage2_desc_handle, ALF_TASK_DESC_ACCEL_KERNEL_REF_L, (unsigned long)app_kernel_name_2);
  alf_task_desc_set_int64(stage2_desc_handle, ALF_TASK_DESC_ACCEL_CTX_SETUP_REF_L, (unsigned long)app_ctx_setup_name_2);
  alf_task_desc_set_int64(stage2_desc_handle, ALF_TASK_DESC_ACCEL_CTX_MERGE_REF_L, (unsigned long)app_ctx_merge_name_2);
  alf_task_desc_set_int64(stage2_desc_handle, ALF_TASK_DESC_ACCEL_INPUT_DTL_REF_L, (unsigned long)app_prep_in_name_2);
  alf_task_desc_set_int64(stage2_desc_handle, ALF_TASK_DESC_ACCEL_OUTPUT_DTL_REF_L, (unsigned long)app_prep_out_name_2);

  // these are app specific
  alf_task_desc_set_int32(stage2_desc_handle, ALF_TASK_DESC_TSK_CTX_SIZE, sizeof(stage2_ctx_t));
  alf_task_desc_set_int32(stage2_desc_handle, ALF_TASK_DESC_WB_PARM_CTX_BUF_SIZE, sizeof(my_wb_parms_t));
  // stage2 is in only
  alf_task_desc_set_int32(stage2_desc_handle, ALF_TASK_DESC_WB_IN_BUF_SIZE, BLOCK_SIZE*sizeof(alf_data_int32_t));
  alf_task_desc_set_int32(stage2_desc_handle, ALF_TASK_DESC_WB_OUT_BUF_SIZE, 0);
  alf_task_desc_set_int32(stage2_desc_handle, ALF_TASK_DESC_WB_INOUT_BUF_SIZE, 0);
  alf_task_desc_set_int32(stage2_desc_handle, ALF_TASK_DESC_NUM_DTL_ENTRIES, BLOCK_SIZE/4096);
  alf_task_desc_set_int32(stage2_desc_handle, ALF_TASK_DESC_NUM_DTL, 1);
  alf_task_desc_set_int32(stage2_desc_handle, ALF_TASK_DESC_MAX_STACK_SIZE, 1024);
  alf_task_desc_set_int32(stage2_desc_handle, ALF_TASK_DESC_PARTITION_ON_ACCEL, 1);
  // the task context  
  alf_task_desc_ctx_entry_add(stage2_desc_handle, ALF_DATA_INT32, sizeof(my_context_sim_t)/sizeof(alf_data_int32_t));

  
  printf("%s is initializing input.\n", app_name);fflush(stdout);

  /* prepare data for input */
  prepare_data();

  printf("%s is initialized. Computing results...\n", app_name);
  
  rc = alf_dataset_create(alf_handle, &dataset_handle[0]);
  if (rc < 0) {
    fprintf(stderr, "Call alf_dataset_create error: %d\n", rc);
    return 1;
  }
  rc = alf_dataset_create(alf_handle, &dataset_handle[1]);
  if (rc < 0) {
    fprintf(stderr, "Call alf_dataset_create error: %d\n", rc);
    return 1;
  }
  rc = alf_dataset_buffer_add(dataset_handle[0], stage_buffer[0], BLOCK_SIZE*WB_PER_TASK*sizeof(alf_data_int32_t), ALF_DATASET_READ_WRITE);
  if (rc < 0) {
    fprintf(stderr, "Call alf_dataset_buffer_add error: %d\n", rc);
    return 1;
  }
  rc = alf_dataset_buffer_add(dataset_handle[1], stage_buffer[1], BLOCK_SIZE*WB_PER_TASK*sizeof(alf_data_int32_t), ALF_DATASET_READ_WRITE);
  if (rc < 0) {
    fprintf(stderr, "Call alf_dataset_buffer_add error: %d\n", rc);
    return 1;
  }

  old_tsk_handles[0][0] = old_tsk_handles[0][1] =
      old_tsk_handles[1][0] = old_tsk_handles[1][1] = ALF_NULL_HANDLE;  // just to get rid of compiler warnings

  wb_parm.size = BLOCK_SIZE;
  
  for(i=0; i<TOTAL_SIM_TIME/(BLOCK_SIZE*WB_PER_TASK); i++)
  {
    
     rc = alf_task_create(stage1_desc_handle, &stage1_ctx, 1, ALF_TASK_ATTR_SCHED_FIXED, 0, &tsk1);
     if (rc < 0) 
     {
         fprintf(stderr, "**ERR: cannot create stage1 [%d] task(%d).\n", i, rc);
         alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);
         return -1;
     }
     rc = alf_task_create(stage2_desc_handle, &stage2_ctx, 1, ALF_TASK_ATTR_SCHED_FIXED, 0, &tsk2);
     if (rc < 0) 
     {
         fprintf(stderr, "**ERR: cannot create stage1 [%d] task(%d).\n", i, rc);
         alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);
         return -1;
     }
     
     rc = alf_task_dataset_associate(tsk1, dataset_handle[i&1]);
     if (rc < 0) {
       fprintf(stderr, "Call alf_task_dataset_associate error: %d\n", rc);
       return 1;
     }
     
     rc = alf_task_dataset_associate(tsk2, dataset_handle[i&1]);
     if (rc < 0) {
       fprintf(stderr, "Call alf_task_dataset_associate error: %d\n", rc);
       return 1;
     }
     
     // task2[i] need task1[i] to complete
     rc = alf_task_depends_on(tsk2,tsk1);
     if (rc < 0) 
     {
         fprintf(stderr, "**ERR: set_dep %d failed (%d).\n", i, rc);
         alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);
         return -1;
     }
     
     if(i > 0)  // we have previous tasks
     {
         // task1[i] need task1[i-1] to complete because shared task context
         rc = alf_task_depends_on(tsk1, old_tsk_handles[0][(i-1)&1]); 
         if (rc < 0) 
         {
             fprintf(stderr, "**ERR: set_dep %d failed (%d).\n", i, rc);
             alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);
             return -1;
         }
         // task2[i] need task2[i-1] to complete because shared task context
         rc = alf_task_depends_on(tsk2, old_tsk_handles[1][(i-1)&1]); 
         if (rc < 0) 
         {
             fprintf(stderr, "**ERR: set_dep %d failed (%d).\n", i, rc);
             alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);
             return -1;
         }

         if(i > 1)
         {
             // task1[i] need task2[i-2] to complete because shared stage buffer
             rc = alf_task_depends_on(tsk1, old_tsk_handles[1][i&1]);  //(i-2)&1 = i&1 
             if (rc < 0) 
             {
                 fprintf(stderr, "**ERR: set_dep %d failed (%d).\n", i, rc);
                 alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);
                 return -1;
             }
         }
     }
     old_tsk_handles[0][i&1] = tsk1;
     old_tsk_handles[1][i&1] = tsk2;

#ifdef _ALF_64B_  // 64 bit
     wb_parm.addr = (alf_data_addr64_t)stage_buffer[i&1];
#else  // 32 bit
     wb_parm.addr = (alf_data_addr64_t)(alf_data_addr32_t)stage_buffer[i&1];
#endif
      /* we use multi use work block for the stage1 */
      alf_wb_create(tsk1, ALF_WB_MULTI, WB_PER_TASK,  &wb_handle);
      alf_wb_parm_add(wb_handle, (void *)&wb_parm.size, 2, ALF_DATA_INT32, 0); // size and dummy
      alf_wb_parm_add(wb_handle, (void *)&wb_parm.addr, 1, ALF_DATA_ADDR64, 0); // addr
      alf_wb_enqueue(wb_handle);
      alf_task_finalize(tsk1);
     
      /* we use multi use work block for the stage2 */
      alf_wb_create(tsk2, ALF_WB_MULTI,  WB_PER_TASK,  &wb_handle);
      alf_wb_parm_add(wb_handle, (void *)&wb_parm.size, 2, ALF_DATA_INT32, 0); // size and dummy
      alf_wb_parm_add(wb_handle, (void *)&wb_parm.addr, 1, ALF_DATA_ADDR64, 0); // addr
      alf_wb_enqueue(wb_handle);
      alf_task_finalize(tsk2);
      
  }
  

  /* alf task wait until all tasks are processed */
  rc = alf_task_wait(tsk2, -1);
  if(rc < 0)
  {
      fprintf(stderr, "**ERR: task wait failure(%d).\n", rc);
      alf_exit(alf_handle, ALF_EXIT_POLICY_FORCE, 0);
      return -1;
  }
  
  printf("%s is finished. Now verifying results...\n", app_name);

  rc = verify_results();
  if (rc != 0) {
    fprintf(stderr, "**ERR: verification failure(%d).\n", rc);
  } else {
    printf("%s verified successfully\n", app_name);
  }
  
  alf_exit(alf_handle, ALF_EXIT_POLICY_WAIT, -1);

  return rc;
}



