/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include "sim.h"

// ---------------------------------------------
// this is a simple simulation
// ---------------------------------------------

// initialize the simulation
void init_sim(my_context_sim_t *p_ctx)
{
	p_ctx->x = BOX_WIDTH/2;
	p_ctx->y = BOX_HEIGHT/2;
	p_ctx->left  = 0;
	p_ctx->lower = 0;
	p_ctx->right = 0;
	p_ctx->upper = 0;
    p_ctx->total_sims = 0;

	return;
}

// the single step of simulation, rx, ry is two random numbers for movements in x and y direction
unsigned int sim_one_step(my_context_sim_t *p_ctx, unsigned int rx, unsigned int ry)
{
	int mv_x = (int)(rx%(MOVE_MAX*2+1)) - MOVE_MAX;  //make sure -MOVE_MAX <= mv <= MOVE_MAX
	int mv_y = (int)(ry%(MOVE_MAX*2+1)) - MOVE_MAX;
	unsigned int hit = 0;  // if it hit a wall

    // the moving directions 
	switch (((mv_x < 0)?1:0) | ((mv_y < 0)?2:0))
	{
	case 0:  // mv_x >= 0 && mv_y >= 0
        // the wall hit test
		switch(((p_ctx->x + mv_x >= BOX_WIDTH)?1:0) | ((p_ctx->y + mv_y >= BOX_HEIGHT)?2:0))
		{
		case 0: // no one hit walls
			p_ctx->x += mv_x;
			p_ctx->y += mv_y;
			break;
		case 1:  // hit right wall
			hit = 1;
			p_ctx->right ++;
			break;
		case 2:  // hit upper wall
			hit = 1;
			p_ctx->upper ++;
			break;
		case 3: // need to decide which wall is hit firstly
			{
				int dx, dy;
				hit = 1;
				dx = BOX_WIDTH - p_ctx->x;
				dy = BOX_HEIGHT - p_ctx->y;

				if(mv_x/dx >= mv_y/dy)  // hit right early
					p_ctx->right ++;
				else  // hit upper first;
					p_ctx->upper ++;
			}
			break;
		}
		break;

	case 1:  // mv_x <  0 && mv_y >= 0
        // the wall hit test
		switch(((p_ctx->x + mv_x <= 0)?1:0) | ((p_ctx->y + mv_y >= BOX_HEIGHT)?2:0))
		{
		case 0: // no one hit walls
			p_ctx->x += mv_x;
			p_ctx->y += mv_y;
			break;
		case 1:  // hit left wall
			hit = 1;
			p_ctx->left ++;
			break;
		case 2:  // hit upper wall
			hit = 1;
			p_ctx->upper ++;
			break;
		case 3: // need to decide which wall is hit firstly
			{
				int dx, dy;
				hit = 1;
				dx = p_ctx->x;
				dy = BOX_HEIGHT - p_ctx->y;

				if(-mv_x/dx >= mv_y/dy)  // hit left early
					p_ctx->left ++;
				else  // hit upper first;
					p_ctx->upper ++;
			}
			break;
		}
		break;

	case 2:  // mv_x >= 0 && mv_y <  0
        // the wall hit test
		switch(((p_ctx->x + mv_x >= BOX_WIDTH)?1:0) | ((p_ctx->y + mv_y <= 0)?2:0))
		{
		case 0: // no one hit walls
			p_ctx->x += mv_x;
			p_ctx->y += mv_y;
			break;
		case 1:  // hit right wall
			hit = 1;
			p_ctx->right ++;
			break;
		case 2:  // hit lower wall
			hit = 1;
			p_ctx->lower ++;
			break;
		case 3: // need to decide which wall is hit firstly
			{
				int dx, dy;
				hit = 1;
				dx = BOX_WIDTH - p_ctx->x;
				dy = p_ctx->y;

				if(mv_x/dx >= -mv_y/dy)  // hit right early
					p_ctx->right ++;
				else  // hit lower first;
					p_ctx->lower ++;
			}
			break;
		}
		break;

	case 3:  // mv_x <  0 && mv_y <  0
        // the wall hit test
		switch(((p_ctx->x + mv_x <= 0)?1:0) | ((p_ctx->y + mv_y <= 0)?2:0))
		{
		case 0: // no one hit walls
			p_ctx->x += mv_x;
			p_ctx->y += mv_y;
			break;
		case 1:  // hit left wall
			hit = 1;
			p_ctx->left ++;
			break;
		case 2:  // hit lower wall
			hit = 1;
			p_ctx->lower ++;
			break;
		case 3: // need to decide which wall is hit firstly
			{
				int dx, dy;
				hit = 1;
				dx = p_ctx->x;
				dy = p_ctx->y;

				if(-mv_x/dx >= -mv_y/dy)  // hit left early
					p_ctx->left ++;
				else  // hit lower first;
					p_ctx->lower ++;
			}
			break;
		}
		break;
	}

	if(hit)  // need to reset position
	{
		p_ctx->x = BOX_WIDTH/2;
		p_ctx->y = BOX_HEIGHT/2;
	}
    p_ctx->total_sims++;
	return hit;
}


