/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

#include <alf_accel.h>
#include "../pipe_line.h"
#include "../sim.c"       // sorry, but this is easier for me since I share the code
#include "../my_rng.c"    // sorry, but this is easier for me since I share the code 

int rng_kernel(void *p_task_context  ,
                   void *p_parm_context, 
                   void *p_input_buffer   __attribute__ ((unused)), 
                   void *p_output_buffer  ,
                   void *p_inout_buffer  __attribute__ ((unused)), 
                   unsigned int current_count  __attribute__ ((unused)), 
                   unsigned int total_count __attribute__ ((unused)))
{
  my_context_rng_t *p_ctx = (my_context_rng_t *) p_task_context;
  unsigned int i;
  unsigned int *buf;
  my_wb_parms_t *p_parm = (my_wb_parms_t *) p_parm_context;

  buf = (unsigned int *) p_output_buffer;

  for (i = 0; i < p_parm->size; i++)
    buf[i] = lfrng(p_ctx);

  return 0;
}

int rng_prep_out(void *p_task_context __attribute__ ((unused)),
                   void *p_parm_ctx_buffer,
                   void *p_dtl,
                   unsigned int current_count,
                   unsigned int total_count __attribute__ ((unused)))
{
    my_wb_parms_t *p_parm = (my_wb_parms_t *) p_parm_ctx_buffer;
    unsigned int k;
    ALF_ACCEL_DTL_BEGIN(p_dtl, ALF_BUF_OUT, 0);
    for(k=0; k<p_parm->size; k+=4096)  // Cell Specific 16KB issue, 4096*int32 is 16KB
    {
       unsigned int bsize;
       if(k+4096 <= p_parm->size)
           bsize = 4096;
       else
           bsize = p_parm->size - k;
       ALF_ACCEL_DTL_ENTRY_ADD(p_dtl, bsize, ALF_DATA_INT32, p_parm->addr + (current_count*p_parm->size + k)*sizeof(int));
    }
    ALF_ACCEL_DTL_END(p_dtl);
    return 0;
}


int sim_kernel(void *p_task_context  ,
                   void *p_parm_context, 
                   void *p_input_buffer  , 
                   void *p_output_buffer   __attribute__ ((unused)),
                   void *p_inout_buffer  __attribute__ ((unused)), 
                   unsigned int current_count  __attribute__ ((unused)), 
                   unsigned int total_count __attribute__ ((unused)))
{
  my_context_sim_t *p_ctx = (my_context_sim_t *) p_task_context;
  unsigned int i;
  unsigned int *buf;
  my_wb_parms_t *p_parm = (my_wb_parms_t *) p_parm_context;

  buf = (unsigned int *) p_input_buffer;

  for (i = 0; i < p_parm->size; i+=2)
    sim_one_step(p_ctx, buf[i], buf[i+1]);

  return 0;
}

int sim_prep_in(void *p_task_context __attribute__ ((unused)),
                   void *p_parm_ctx_buffer,
                   void *p_dtl,
                   unsigned int current_count,
                   unsigned int total_count __attribute__ ((unused)))
{
    my_wb_parms_t *p_parm = (my_wb_parms_t *) p_parm_ctx_buffer;
    unsigned int k;
    ALF_ACCEL_DTL_BEGIN(p_dtl, ALF_BUF_IN, 0);
    for(k=0; k<p_parm->size; k+=4096)  // Cell Specific 16KB issue, 4096*int32 is 16KB
    {
       unsigned int bsize;
       if(k+4096 <= p_parm->size)
           bsize = 4096;
       else
           bsize = p_parm->size - k;
       ALF_ACCEL_DTL_ENTRY_ADD(p_dtl, bsize, ALF_DATA_INT32, p_parm->addr + (current_count*p_parm->size + k)*sizeof(int));
    }
    ALF_ACCEL_DTL_END(p_dtl);
    return 0;
}

int context_merge(void* p_dst_task_ctx __attribute__ ((unused)), 
                  void* p_task_ctx __attribute__ ((unused)))
{
  //Nothing needed since there's only one accelerator per task
  return 0;
}


ALF_ACCEL_EXPORT_API_LIST_BEGIN
   ALF_ACCEL_EXPORT_API("rng_kernel", rng_kernel);
   ALF_ACCEL_EXPORT_API("rng_prep_out", rng_prep_out);
   ALF_ACCEL_EXPORT_API("sim_kernel", sim_kernel);
   ALF_ACCEL_EXPORT_API("sim_prep_in", sim_prep_in);
   ALF_ACCEL_EXPORT_API("context_merge", context_merge);
ALF_ACCEL_EXPORT_API_LIST_END
