/* --------------------------------------------------------------  */
/* (C)Copyright 2007,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

/***************************************************************
*   Accelerator side code
***************************************************************/

#include <stdio.h>
#include <alf_accel.h>
#include "../min_max.h"

/* the computation kernel function */
int comp_kernel(void *p_task_context,
                    void *p_parm_ctx_buffer,
                    void *p_input_buffer,
                    void *p_output_buffer __attribute__ ((unused)),
                    void *p_inout_buffer __attribute__ ((unused)),
                    unsigned int current_count  __attribute__ ((unused)),
                    unsigned int total_count __attribute__ ((unused)))
{
    my_task_context_t *p_ctx = (my_task_context_t *) p_task_context;
    my_wb_parms_t *p_parm = (my_wb_parms_t *) p_parm_ctx_buffer;

    alf_data_int32_t *a = (alf_data_int32_t *)p_input_buffer;
    unsigned int size = p_parm->num_data;
    unsigned int i;

    // I won't do any vector optimization here, for readability
    for(i=0;i<size;i++) {
        if(a[i]>p_ctx->max) 
            p_ctx->max = a[i];
        else if(a[i]<p_ctx->min) 
            p_ctx->min = a[i];        
    } 
    return 0;
}


/* the context merge function */

int ctx_merge(void* p_task_context_to_be_merged, void* p_task_context)
{
    my_task_context_t *p_ctx = (my_task_context_t *) p_task_context;
    my_task_context_t *p_mgr_ctx = (my_task_context_t *) p_task_context_to_be_merged;

    //printf("merge thread %d: (mmin %d mmax %d) -> (min %d max %d)\n",alf_accel_instance_id(), p_mgr_ctx->min, p_mgr_ctx->max, p_ctx->min, p_ctx->max);
    if(p_mgr_ctx->max > p_ctx->max)
        p_ctx->max = p_mgr_ctx->max;
    if(p_mgr_ctx->min < p_ctx->min)
                p_ctx->min = p_mgr_ctx->min;
    return 0;
}


// in this example, this setup API is not necessary as the context has been initialized by the main app
// and we do not need thread specific initial context
int ctx_setup(void* p_task_context __attribute__ ((unused)))
{
   //my_task_context_t *p_ctx = (my_task_context_t *) p_task_context;
   //printf("setup thread %d: (min %d max %d)\n",alf_accel_instance_id(), p_ctx->min, p_ctx->max);
   return 0;
}
/* export the symbols for ALF runtime*/

ALF_ACCEL_EXPORT_API_LIST_BEGIN
   ALF_ACCEL_EXPORT_API("comp_kernel", comp_kernel);
   ALF_ACCEL_EXPORT_API("ctx_merge", ctx_merge);
   ALF_ACCEL_EXPORT_API("ctx_setup", ctx_setup);
ALF_ACCEL_EXPORT_API_LIST_END


